"""
File Purpose: tools related to xarray dimensions
"""
import numpy as np
import xarray as xr
try:
    from xarray.computation.rolling import DataArrayCoarsen
except ImportError:
    from xarray.core.rolling import DataArrayCoarsen
# note: can't use "xr.core.rolling.DataArrayCoarsen" directly in a new python kernel,
#    although it does work after creating an xr.DataArray.coarsen() object.
#    imported it here so that we can use xr.DataArray.coarsen docstring.

from .xarray_accessors import pcAccessor
from .xarray_misc import xarray_copy_kw
from ..docs_tools import format_docstring
from ..sentinels import UNSET
from ...errors import (
    DimensionalityError, DimensionValueError, DimensionKeyError,
    InputConflictError,
)


''' --------------------- Docstrings --------------------- '''

_paramdocs_ensure_dims = {
    'promote_dims_if_needed': '''bool
        whether to promote non-dimension coords to dimensions.
        if False, raise DimensionKeyError if any relevant coord is not already a dimension.''',
    'missing_dims': '''str in ('ignore', 'warn', 'raise')
        what to do if any coord is not found:
            'ignore' --> do nothing.
            'warn' --> raise a warning.
            'raise' --> raise DimensionKeyError.''',
}


''' --------------------- Dimensions --------------------- '''

def is_iterable_dim(value, *, min_length=None):
    '''returns whether value represents multiple values (of a dimension).

    if value.ndim exists,
        return False if ndim==0,
        else True if ndim > 0,
        else raise DimensionalityError.
    # else, return whether iter(value) succeeds.  # <-- no longer used; use is_scalar from xr.
    else, return (not xr.core.utils.is_scalar(value))

    min_length: None or int
        if provided, before returning True, require that len(value) >= min_length.
    '''
    if hasattr(value, 'ndim'):
        ndim = value.ndim
        if ndim == 0:
            result = False
        elif ndim > 0:
            result = True
        else:
            raise DimensionalityError(f"expected ndim >= 0, but got ndim={ndim}.")
    else:
        result = (not xr.core.utils.is_scalar(value))
    if result and (min_length is not None):
        try:
            L = len(value)
        except TypeError:
            result = False  # couldn't determine length, so assume it's not long enough.
        else:
            result = result and (L >= min_length)
    return result

def take_along_dimension(dimension, array, at=None, *, i=None, default=UNSET,
                         drop_labels=False, as_dict=False, item=False):
    '''returns something like: [array.sel({dimension: val}) for val in ``at``]
    i.e., list of taking these values along this dimension of array.

    at: None or list-like of values in this dimension
        take at these values. None --> use values from array.coords[dimension]
    i: None or indices.
        (if provided) take only at these indices; use isel.
    default: UNSET or any value
        if provided, use this value for any at[i] not in array.coords[dimension].
        if not already an xarray.DataArray, convert it, and set coords[dimension]=at[i].
            (e.g. take_along_dimension('component', arr, at=['x'], default=0),
            if arr doesn't have 'component'=='x', will give xarray.DataArray(0, coords={'component': 'x'}))
    drop_labels: bool, default False
        if True, drop the labels along the taken dimension.
        E.g., if dimension=='component' and labels=['x', 'y'],
            by default, result[0].component=='x' and result[1].component=='y'
            but if drop_labels then result[0] and result[1] will not have .component at all.
    as_dict: bool, default False
        if True, return dict of {dim value: array value at this dim}
            (instead of a list of array value at each dim value)
    item: bool
        if True, convert arrays to single values via array.item().

    if array is 0-d along dimension,
        returns [array] if labels corresponds to the one label in this dimension.
    '''
    if i is not None:
        if at is not None:
            raise InputConflictError('cannot provide both "at" and "i".')
        iseled = array.isel({dimension: i})
        return take_along_dimension(dimension, iseled, drop_labels=drop_labels, as_dict=as_dict)
    # else, i is None
    coords = array.coords[dimension]
    # 0d coords
    if coords.ndim == 0:
        coord = coords.item()
        if at is None:
            result = [array]
            used_at = [coord]
        else:
            if all(val == coord for val in at):
                result = [array] * len(at)  # list multiplication
                used_at = at
            else:
                errmsg = (f'provided value for "at" is incompatible with 0-d {dimension!r} values; '
                          f'\ncoords={coords};\nat={at}')
                raise DimensionalityError(errmsg)
    # 1d coords
    elif coords.ndim == 1:
        if at is None:
            # use all of them.
            result = [array.isel({dimension: i_}) for i_ in range(len(coords))]
            used_at = [val.item() for val in coords]
        else:
            # use only some of them.
            # [TODO] xarray.sel doesn't really like DimensionValue... why not?
            #   (using another way to do it, in the meantime.)
            result = []
            for val in at:
                i_ = np.nonzero((coords == val).values)[0]
                if len(i_) == 0:
                    if default is UNSET:
                        raise DimensionValueError(f'no match for val={val!r} in coords={coords}')
                    else:
                        if isinstance(default, xr.DataArray):
                            result.append(default)
                        else:
                            result.append(xr.DataArray(default, coords={dimension: val}))
                elif len(i_) == 1:
                    result.append(array.isel({dimension: i_[0]}))
                else:
                    raise DimensionValueError(f'expected 1 match, but got {len(i_)} matches for val={val!r}')
            used_at = at
    # 2d+ coords
    else:
        raise NotImplementedError(f'take_along_dimension {dimension!r} with ndim={coords.ndim}')
    # postprocessing / bookkeping
    if drop_labels:
        result = [array.drop_vars(dimension) for array in result]
    if item:
        result = [array.item() for array in result]
    if as_dict:
        result = {val: array for val, array in zip(used_at, result)}
    return result

@pcAccessor.register
def take_along_dimensions(array, dimensions, *, atleast_1d=False, **kw_isel):
    '''returns result of taking array along each of these dimensions, in order.
    result will be a numpy array with dtype=object, shape=(d0, d1, ...),
        where di = len(array.coords[dimensions[i]]).
    each element of result will be an xarray.DataArray.

    any dimension can be None --> result shape will be 1 at that dimension, and nothing will be taken.
        E.g. take_along_dimensions(array, [None, 'fluid']) gives array of shape (1, len(fluids)).

    if dimensions is an empty list:
        result will be a scalar numpy array with result[()] = `array`,
        unless atleast_1d=True; then result has shape (1,) and result[0] = `array`.
    '''
    if isinstance(dimensions, str):
        dimensions = [dimensions]
    shape = tuple((1 if dim is None else len(array.coords[dim])) for dim in dimensions)
    result = np.empty(shape, dtype=object)
    if len(shape) == 0:
        if atleast_1d:
            result = np.empty((1,), dtype=object)
            result[0] = array
        else:
            result[()] = array
    else:
        for idx in np.ndindex(*shape):
            selecting = {dim: i for dim, i in zip(dimensions, idx) if dim is not None}
            result[idx] = array.isel(selecting, **kw_isel)
    return result

def join_along_dimension(dimension, arrays, labels=None, *, coords='minimal', **kw_xarray_concat):
    '''returns xr.concat(arrays, dimension). if len(arrays) == 1, instead return arrays[0], unchanged.
    if labels is provided, set result[dimension] = labels
    '''
    if len(arrays) == 1:
        return arrays[0]
    result = xr.concat(arrays, dimension, coords=coords, **kw_xarray_concat)
    if labels is not None:
        result[dimension] = labels
    return result

@pcAccessor.register('rename')
def xarray_rename(array, names=None, **more_names):
    '''return array.rename(names, **more_names), but skip any names not found in array.coords.
    names should be a dict, if provided.
    '''
    if names is not None:
        more_names.update(names)
    apply_names = {name: val for name, val in more_names.items() if name in array.coords}
    return array.rename(apply_names)

@pcAccessor.register('assign')
def xarray_assign(array, coords=None, attrs=None, *, name=UNSET, overwrite=None, expand_if_iterable=False):
    '''array.assign_coords(coords).assign_attrs(attrs).rename(name)
    
    coords: None or dict of {dim: coord}
        each coord must be "non-iterable", as per is_iterable_dim(),
        unless expand_if_iterable=True.
    attrs: None or dict
        assign these attrs. dict of arbitrary values.
    name: UNSET, None, or str
        assign this name to the result, if provided.
    overwrite: None or bool
        whether to overwrite an existing value for coord in array.
        (note - array will never be altered here; only the result might be altered.)
        If any coord already in array.coords, behavior depends on overwrite:
            None --> crash with DimensionKeyError.
            True --> overwrite the coord using the new value.
            False --> return array, unchanged.
    expand_if_iterable: bool
        whether to expand_dims for any iterable coords,
        e.g. array.expand_dims(coords) for the relevant coords.
    '''
    array0 = array  # helps with debugging
    if coords is not None:
        iterable_coords = {dim: val for dim, val in coords.items() if is_iterable_dim(val)}
        if iterable_coords:
            if expand_if_iterable:
                array = array.expand_dims(iterable_coords)
                coords = {dim: val for dim, val in coords.items() if dim not in iterable_coords}
            else:
                errmsg = (f'cannot assign iterable coord(s) when expand_if_iterable=False;'
                          f'got these iterable coords: {iterable_coords}')
                raise DimensionalityError(errmsg)
        if not overwrite:
            coords_already_assigned = set(array.coords).intersection(coords)
            if coords_already_assigned:
                if overwrite is None:
                    errmsg = (f'cannot assign already-assigned coords: {coords_already_assigned}, '
                              f'when overwrite=None.\nTo disable this error, use '
                              f'overwrite=True to update existing coords, or False to skip existing coords.')
                    raise DimensionKeyError(errmsg)
                else:
                    coords = {dim: val for dim, val in coords.items() if dim not in coords_already_assigned}
        if len(coords) > 0:
            array = array.assign_coords(coords)
    if attrs is not None:
        array = array.assign_attrs(attrs)
    if name is not UNSET:
        array = array.rename(name)
    return array

@pcAccessor.register('promote_dim')
def xarray_promote_dim(array, coord, *coords_as_args):
    '''Promote this coord (or these coords) to be a dimension, if it isn't already.
    
    coord: str or iterable of strs
        name of coord(s) to promote.
        if already in array.dims, do nothing.
        if 0D, array.expand_dims(coord).
                (This occurs when coord has no associated dimension, in array.)
        if 1D, array.swap_dims(dict(dim=coord)),
                where dim is the dimension associated with coord.
        if 2D+, crash with DimensionalityError.

    returns array, or a copy of array where coord is one of the dimensions.
    '''
    coords = [coord] if isinstance(coord, str) else coord
    coords = list(coords) + list(coords_as_args)
    to_expand = []
    to_swap = {}
    for coord in coords:
        if coord in array.dims:
            continue  # no need to promote this dim.
        c = array.coords[coord]
        cdims = c.dims
        if len(cdims) == 0:
            to_expand.append(coord)
        elif len(cdims) == 1:
            d = cdims[0]
            if d in to_swap:
                errmsg = (f'multiple coords ({to_swap[d]!r}, {coord!r}) with same dimension ({d!r}), '
                            'cannot promote both of them at the same time.')
                raise DimensionKeyError(errmsg)
            to_swap[d] = coord
        else:
            errmsg = f'cannot promote_dim(coord={coord!r}) for coord with >1 dims: {cdims}'
            raise DimensionalityError(errmsg)
    if 0 == len(to_expand) == len(to_swap):
        return array  # nothing to change.
    if to_expand:
        array = array.expand_dims(tuple(to_expand))
    if to_swap:
        array = array.swap_dims(to_swap)
        # NOTE: swapping when to_swap is empty SHOULD probably just make a copy of array,
        #    however it fails (in xarray version 2024.7.0, at least) to maintain MultiIndex properly,
        #    leading to 'ValueError: cannot re-index or align objects with conflicting indexes...'
        #    with MultiIndex conflicting indexes all completely unrelated to the to_expand and to_swap here,
        #    when doing array math with some promote_dim'd arrays and some non-promote_dim'd arrays.
        # (Workaround: use an 'if' block, so we only swap dims if there are dims to swap.
        #    This workaround might not work always, but it at least prevents the issue when to_swap is empty.)
    return array

@pcAccessor.register('ensure_dims')
@format_docstring(**_paramdocs_ensure_dims)
def xarray_ensure_dims(array, coords, *,
                       promote_dims_if_needed=True, missing_dims='raise',
                       assert_1d=False, return_existing_dims=False):
    '''return array but ensure these coords are dimensions.

    coords: str or iterable of strings
        coords to ensure are dimensions.
    promote_dims_if_needed: {promote_dims_if_needed}
            0D coord --> array.expand_dims(coord)
            1D coord --> array.swap_dims(dict(dim=coord)) for associated dim
            2D+ coord --> crash with DimensionalityError.
    missing_dims: {missing_dims}
    assert_1d: bool, default False
        whether to assert that each of these coords is 1D (after promoting if needed).
    return_existing_dims: bool, default False
        True --> returns [array, set of dims (from input coords) which actually exist]
        probably only useful if missing_dims != 'raise'.
    '''
    if isinstance(coords, str):
        coords = [coords]
    found_missing_dims = set()
    # promote coords
    for cname in coords:
        if promote_dims_if_needed and cname in array.coords:
            array = xarray_promote_dim(array, cname)
        if cname not in array.dims:
            found_missing_dims.add(cname)
            continue
        if assert_1d:
            c = array.coords[cname]
            if c.ndim != 1:
                errmsg = f'ensure_dims expected 1D coord={cname!r}, but got ndim={c.ndim}.'
                raise AssertionError(errmsg)
    # handle missing dims
    if len(found_missing_dims) > 0:
        if missing_dims not in ('ignore', 'warn', 'raise'):
            errmsg = f'invalid missing_dims={missing_dims!r}. Expected "ignore", "warn" or "raise".'
            raise InputError(errmsg)
        if missing_dims =='raise' or missing_dims == 'warn':  # define the error message
            c_or_d = 'coords' if promote_dims_if_needed else 'dims'
            errmsg = (f'Dimensions {found_missing_dims} not found in '
                      f'array.{c_or_d}={set(getattr(array, c_or_d))},\n'
                      f'and missing_dims={missing_dims!r} (using "ignore" would ignore this instead).')
            if missing_dims == 'raise':
                raise DimensionKeyError(errmsg)
            elif missing_dims == 'warn':
                warnings.warn(errmsg)
    # return result
    if return_existing_dims:
        existing_dims = set(coords) - found_missing_dims
        return array, existing_dims
    else:
        return array

@pcAccessor.register('squeeze', totype='array')
def xarray_squeeze(array, dim=None, *, keep=None, drop=True):
    '''return array but drop redundant dims.
    dims are redundant if np.all(array==array.isel(dim=0)).
    This is a more aggressive version of array.squeeze():
        all dims with size 1 are trivially redundant and will be dropped;
        but here, additionally, all dims which are redundant in any way will be dropped.

    dim: None, str, or list of strs
        if provided, only consider dropping these dims.
    keep: None, str, or list of strs
        if provided, do not consider dropping these dims.
        (can provide dim or keep, but not both.)
    drop: bool
        whether to drop coord for size-1 redundant dims.
        if False, will keep scalar coord for size-1 redundant dims.

    E.g. [[0,1,2],[0,1,2]] --> [0,1,2]. (But, as xarray, so result remains well-labeled.)
    '''
    if keep is not None and dim is not None:
        raise InputConflictError(f'cannot provide both dim & keep. Got dim={dim!r}, keep={keep!r}')
    if keep is None:
        dims = array.dims if dim is None else ([dim] if isinstance(dim, str) else dim)
        if any(d not in array.dims for d in dims):
            raise DimensionKeyError(f'all dims must exist, but got dims={dims!r}, array.dims={array.dims}')
    else:
        if isinstance(keep, str): keep = [keep]
        dims = set(array.dims) - set(keep)
    for d in dims:
        dropped = array.isel({d: 0}, drop=drop)
        # could just check np.all(dropped == array), but let's do a few smaller checks first,
        #    to save some computation time in case of large arrays.
        if array.sizes[d] == 1:  # trivially redundant
            array = dropped
        elif np.any(dropped != array.isel({d: 1})):  # clearly not redundant
            continue
        elif np.all(dropped == array):
            array = dropped
    return array

@pcAccessor.register('squeeze_close', totype='array')
def xarray_squeeze_close(array, dim=None, tol=1e-3, *, closeness_name='closeness_{dim}', keep=None, drop=True):
    '''return array but drop redundant dims, with an tolerance when checking equality.
    dims are redundant if all values are close along that dim.
        Closeness check here is: (array.isel(dim=0)/array).std() < tol.
    For each removed dim:
        result will just use array.isel(dim=0).
        result will add coord 'closeness_{dim}' == std(array.isel(dim=0)/array).
    This is a much more aggressive version of array.squeeze():
        all dims with size 1 are trivially redundant and will be dropped;
        but here, additionally, all dims which are redundant in any way will be dropped.
        NOTE: performs all closeness checks before dropping any dims.

    dim: None, str, or list of strs
        if provided, only consider dropping these dims.
    tol: number
        drops dim with tol > (array.isel(dim=0)/array).std().
    closeness_name: None or str, default 'closeness_{dim}'
        add new coord with this name, with value == (array.isel(dim=0)/array).std(),
        for each dim dropped. (smaller is closer.) (always 0 for dims of size 1.)
        (closeness_{dim} value tells the pre-squeeze variation along that dim.)
        None --> do not add this coord.
    keep: None, str, or list of strs
        if provided, do not consider dropping these dims.
        (can provide dim or keep, but not both.)
    drop: bool
        whether to drop coord for size-1 redundant dims.
        if False, will keep scalar coord for size-1 redundant dims.

    See also: xarray_closeness
    '''
    # [TODO] encapsulate code from xarray_squeeze instead of copy-pasting
    if keep is not None and dim is not None:
        raise InputConflictError(f'cannot provide both dim & keep. Got dim={dim!r}, keep={keep!r}')
    if keep is None:
        dims = array.dims if dim is None else ([dim] if isinstance(dim, str) else dim)
        if any(d not in array.dims for d in dims):
            raise DimensionKeyError(f'all dims must exist, but got dims={dims!r}, array.dims={array.dims}')
    else:
        if isinstance(keep, str): keep = [keep]
        dims = set(array.dims) - set(keep)
    to_drop = []  # (handle all dropping at the end)
    for d in dims:
        dropped = array.isel({d: 0})
        if array.sizes[d] == 1:  # trivially redundant
            closeness = 0
            to_drop.append(d)
        else:
            closeness = (dropped / array).std()
            if closeness < tol:
                to_drop.append(d)
            else:
                continue
        if closeness_name is not None:
            closeness_coord = closeness_name.format(dim=d)
            array = array.assign_coords({closeness_coord: closeness})
    if len(to_drop) > 0:
        array = array.isel({d: 0 for d in to_drop}, drop=drop)
    return array

@pcAccessor.register('closeness', totype='array')
def xarray_closeness(array, dim=None):
    '''computes "closeness" along each dim: (array.isel(dim=0)/array).std()
    Result is a dataset with one scalar data_var telling closeness of each dim.
    Note: the std() is always taken across ALL dims of the array.
        [TODO] allow to keep some dims?

    dim: None, str, or list of strs
        if provided, only consider these dims.

    See also: xarray_squeeze_close
    '''
    if dim is None:
        dims = array.dims
    elif isinstance(dim, str):
        dims = [dim]
    else:
        dims = dim
    result = {}
    for d in dims:
        dropped = array.isel({d: 0})
        closeness = (dropped / array).std()
        result[d] = closeness
    copy_kw = xarray_copy_kw(array, array_to_dataset=True)
    return xr.Dataset(result, **copy_kw)

@pcAccessor.register('drop_unused_dims', totype='dataset')
def xarray_drop_unused_dims(dataset, only=None):
    '''dataset.drop_dims(dims which do not appear in any of the data_vars).
    only: None, str, or list of str
        if provided, only consider dropping dims in this list.
    '''
    used_dims = set()
    for v in dataset.data_vars:
        used_dims.update(dataset[v].dims)
    unused_dims = set(dataset.dims) - used_dims
    if only is not None:
        if isinstance(only, str): only = [only]
        unused_dims &= set(only)
    return dataset.drop_dims(unused_dims)

@pcAccessor.register('drop_vars', totype='dataset')
def xarray_drop_vars(dataset, names, *, errors='ignore', drop_unused_dims=True):
    '''dataset.drop_vars(names, errors=errors), then drop any unused dims.
    errors: 'ignore' or 'raise'
        what to do if any name not found in dataset.
        (passed directly to dataset.drop_vars)
    drop_unused_dims: bool
        whether to also xarray_drop_unused_dims for dims from dropped vars.
    '''
    if isinstance(names, str): names = [names]
    result = dataset.drop_vars(names, errors=errors)
    if drop_unused_dims:
        dropped_var_dims = set()
        for name in names:
            dropped_var_dims.update(dataset[name].dims)
        result = xarray_drop_unused_dims(result, only=dropped_var_dims)
    return result

@pcAccessor.register('popped', totype='dataset')
def xarray_popped(ds, var):
    '''returns (ds[var], copy of ds without var)
    the copy of ds without var also drops any now-unused dims.
    '''
    return (ds[var], xarray_drop_vars(ds, var, drop_unused_dims=True))


''' --------------------- Broadcasting --------------------- '''

@pcAccessor.register('broadcastable', totype='array')
def xarray_broadcastable_array(array, dims):
    '''return broadcastable version of array, standardizing dims to the list provided.
    missing dims will be expanded to size 1. result dims will be put in this order.

    dims: list of str
        result will have all these dims, in this order, with size 1 if not present in array.
        if any array.dims not present in dims, raise DimensionKeyError.
    '''
    extra_dims = set(array.dims) - set(dims)  # dims in array
    if extra_dims:
        errmsg = (f'array not broadcastable to input dims={dims}, '
                  f'due to {extra_dims} in array.dims but not input dims.')
        raise DimensionKeyError(errmsg)
    missing_dims = tuple(set(dims) - set(array.dims))
    if missing_dims:
        array = array.expand_dims(missing_dims)
    return array.transpose(*dims)

@pcAccessor.register('broadcastable', totype='dataset')
def xarray_broadcastable_from_dataset(dataset, var=None, *, dims=None):
    '''return dict of broadcastable versions of data_var(s) from dataset, standardizing data vars' dims.
    missing dims will be expanded to size 1. result's dims will be put in order.

    var: None, str, or list of str.
        str --> return a broadcastable version of this data_var. Result is an xarray.DataArray.
        list of str --> return dict of {v: broadcastable version of v} across v in var.
        None --> use var = list(dataset.keys())
    dims: None or list of str
        result will have all these dims, in this order, with size 1 if not present in array.
        if any array.dims not present in dims, raise DimensionKeyError.
        None --> use dataset.dims.
    '''
    if dims is None:
        dims = dataset.dims
    if isinstance(var, str):
        return xarray_broadcastable_array(dataset[var], dims)
    # else, return dict of {v: broadcastable version of v} across v in var.
    result = {}
    if var is None:
        var = list(dataset.keys())
    for v in var:
        result[v] = xarray_broadcastable_array(dataset[v], dims)
    return result

@pcAccessor.register('from_broadcastable')
def xarray_from_broadcastable(array, broadcastable, *, dims=None, squeeze=True):
    '''return xarray from broadcastable values, using dims/coords/attrs from input.

    array: xarray.DataArray or xarray.Dataset
        read relevant dims, coords, and attrs from this array; copy to result.
    broadcastable: array (possibly numpy array) or dict of arrays
        result's data comes from this array(s).
        single array --> result.data = array.
        dict of arrays --> result.data = array[v] for v in array.  # [TODO] not yet implemented.
    dims: None or list of str
        list of dim names for broadcastable's dims. len(dims) == broadcastable.ndim.
        None --> use array.dims.
    squeeze: bool
        whether to squeeze away dims with size 1 in broadcastable.
        True --> result.dims will only include dims with size > 1 in broadcastable.
    '''
    if dims is None:
        dims = array.dims
    if isinstance(broadcastable, dict):
        raise NotImplementedError('xarray_from_broadcastable not yet implemented for dict inputs.')
    # else, broadcastable is a single array.
    shape = broadcastable.shape
    if len(dims) != len(shape):
        errmsg = f'length of broadcastable shape ({shape}) != length of input dims ({dims}).'
        raise DimensionalityError(errmsg)
    relevant_dims = [d for d, s in zip(dims, shape) if s > 1]
    kw = xarray_copy_kw(array, dims=relevant_dims, name=False)
    if squeeze:
        broadcastable = np.squeeze(broadcastable)
    else:
        kw['dims'] = dims  # using all dims not just relevant dims.
    return xr.DataArray(broadcastable, **kw)



''' --------------------- Coarsen / windowing --------------------- '''

_coarsen_doc = xr.DataArray.coarsen.__doc__
_construct_doc = DataArrayCoarsen.construct.__doc__
if 'Examples\n' in _coarsen_doc:  # gives docstring with Examples removed
    _coarsen_doc = _coarsen_doc[:_coarsen_doc.index('Examples\n')].rstrip()
if 'Examples\n' in _construct_doc:  # gives docstring with Examples removed
    _construct_doc = _construct_doc[:_construct_doc.index('Examples\n')].rstrip()

@pcAccessor.register('coarsened')
@format_docstring(coarsen_doc=_coarsen_doc, construct_doc=_construct_doc)
def xarray_coarsened(array, dim, window_len, dim_coarse='window', dim_fine=None, *,
                     assign_coarse_coords=False,
                     # kw for coarsen:
                     boundary=UNSET, side=UNSET,
                     # kw for construct:
                     stride=UNSET, fill_value=UNSET, keep_attrs=UNSET,
                     ):
    '''construct a coarsened version of array, where dim is coarsened by window_len,
    and becomes two dims: dim_coarse and dim_fine.
    Original dim coords will be associated with dim_coarse and dim_fine in the new array.

    dim: str
        dimension to coarsen.
        if a non-dimension coordinate, will attempt to promote it to a dimension (e.g. via swap_dims).
    window_len: int
        length of the window to coarsen over.
    dim_coarse: str, default 'window'
        name of coarse dimension; the i'th value here corresponds to the i'th window.
    dim_fine: None or str
        name of fine dimension; the j'th value here corresponds to the j'th element within a window.
        if None, use '_'+dim, e.g. dim='t' --> dim_fine='_t'.
    assign_coarse_coords: bool or coords
        coords to assign along the dim_coarse dimension.
        True --> use np.arange.
        False --> don't assign coords.
    boundary, side: UNSET or value
        if provided (not UNSET), pass this value to coarsen().
        boundary should be 'exact', 'trim', or 'pad'.
        side should be 'left' or 'right'.
    stride, fill_value, keep_attrs: UNSET or value
        if provided (not UNSET), pass this value to construct().
    
    docs for coarsen and construct are copied below, for convenience:

    xarray.DataArray.coarsen
    ------------------------
    {coarsen_doc}


    xr.core.rolling.DataArrayRolling.construct
    ------------------------------------------
    {construct_doc}
    '''
    # bookkeeping
    kw_coarsen = dict(boundary=boundary, side=side)
    kw_construct = dict(stride=stride, fill_value=fill_value, keep_attrs=keep_attrs)
    for kw in kw_coarsen, kw_construct:
        for key, val in tuple(kw.items()):
            if val is UNSET:
                del kw[key]
    if dim_fine is None:
        dim_fine = f'_{dim}'
    # promote non-coordinate dim if necessary, else returns array unchanged.
    arr = xarray_promote_dim(array, dim)
    # coarsen & reconstruct
    coarse = arr.coarsen({dim: window_len}, **kw_coarsen)
    result = coarse.construct({dim: (dim_coarse, dim_fine)}, **kw_construct)
    # bookkeeping
    if assign_coarse_coords is not False:
        if assign_coarse_coords is True:
            assign_coarse_coords = np.arange(len(result[dim_coarse]))
        result = result.assign_coords({dim_coarse: assign_coarse_coords})
    return result




"""
File Purpose: DimensionHaver
"""
import contextlib
import itertools

import numpy as np
import xarray as xr

from .dim_point import DimPoint, DimRegion
from .dimension import Dimension
from ..behavior import BehaviorHaver
from ...errors import QuantCalcError, DimensionalityError, InputError
from ...tools import (
    simple_property,
    memory_size_check_loading_arrays_like,
    repr_simple,
    format_docstring,
    alias,
    using_attrs, maintaining_attrs,
    TaskArray, TaskContainerCallKwargsAttrHaver,
)
from ...tools.multiprocessing import _paramdocs_tasks
from ...defaults import DEFAULTS



''' --------------------- DimensionHaver --------------------- '''

_paramdocs_loading = {
    'loader': '''callable of (*args_loader, **kw_loader) -> xarray.DataArray.
        will call loader to get result values at each combination of dims values in self.
        (loader will probably depend on dims values from self.)''',
    'dims': '''iterable of strs or Dimension objects
        load across these Dimensions.
        loads across the current values (when this method was called) of each dimension,
            not necessarily "all" values. (e.g., self.snap, not self.snaps)
        str values --> use self.dimensions[d] (where d is a str in dims).
        len(dims)==0 --> just return loader(var, *args_loader, **kw_loader).
        While loading, set dim.loading=True for each dim.''',
    'assign_coords': '''None or bool, default None
        whether to dim.assign_coord for each result of loader, for each dimension.
        None --> assign coord only if dim.name not already in array.coords.''',
    'array_MBmax': f'''UNSET, None, or number
        maximum result size allowed, in Megabytes.
        will raise a MemorySizeError if result size would be larger than this.
        UNSET --> use DEFAULTS.ARRAY_MBYTES_MAX (default: {DEFAULTS.ARRAY_MBYTES_MAX} MB).
        None --> no limit.
        Assumes that each result (at each dimpoint) will be the same size.''',
}

class DimensionHaver(BehaviorHaver, TaskContainerCallKwargsAttrHaver):
    '''class which can have multiple Dimensions attached to it.

    Manages multiple Dimensions and provides methods for working with them, such as:
        current_n_dimpoints, dim_values, enumerate_dimpoints, get_first_dimpoint.
    Additionally, provides load_across_dims, which is useful when loading arrays from multiple files,
        e.g. see EppicBasesLoader, EbysusBasesLoader.
    The logic for managing the dimensions is implemented in __init_subclass__;
        also note the Dimensions may attach methods to subclasses of this class via setup_haver, e.g.:
            class FluidDimension(Dimension, name='fluid'):
                pass   # no special methods for this example

            @FluidDimension.setup_haver  # setup FluidHaver by attaching various relevant properties & methods.
            class FluidHaver(DimensionHaver, dimension='fluid'):
                pass   # no special methods for this example

    when defining a subclass, can provide these kwargs (e.g. class Foo(DimensionHaver, dimension=...)):
        dimension: str or None
            name of the dimension which has a current value associated with it. e.g. "fluid", "component"
        dim_plural: str or None
            plural name for dimension. if None, append 's' to dimension. e.g. "fluids", "components"
    '''
    # # # PROPERTIES DEFINED HERE # # #
    dimensions = simple_property('_dimensions', setdefaultvia='_get_dimensions_dict',
            doc='''dict of dimensions in self; {dimension name: Dimension object}.
            e.g. {'fluid': self.fluid_dim, 'snap': self.snap_dim, ...}.''')
    def _get_dimensions_dict(self):
        '''return dict of dimensions in self; {dimension name: Dimension object}'''
        return {dim: getattr(self, f'{dim}_dim') for dim in self._dim_types}

    # # # # __INIT_SUBCLASS__ # # # #
    def __init_subclass__(cls, *, dimension=None, dim_plural=None, **kw_super):
        '''called when subclassing DimensionHaver; sets some useful attributes related to dimension.

        dimension: str or None
            name of dimension associated with this subclass.
            if None, no particular dimension associated with this subclass.
        dim_plural: str or None
            plural form of dimension. if None, use str(dimension)+'s'.
    
        Sets various attributes in cls:
            cls._dimension = dimension,
            cls._dim_plural = dim_plural,
            cls._dim_types = dict of all {dimension name: Dimension subclass} from cls and cls.__bases__
                    (note - connecting self._dimension to this dict is not handled here;
                    it is handled by __setup_haver__ called by Dimension.setup_haver)
        '''
        # defaults
        if dim_plural is None:
            dim_plural = None if dimension is None else f'{dimension}s'
        # set attributes
        cls._dimension = dimension
        cls._dim_plural = dim_plural
        cls._dim_types = dict()
        for base in cls.__bases__:
            if hasattr(base, '_dim_types'):
                cls._dim_types.update(base._dim_types)
        # register dimension as a behavior_attr
        if dimension is not None:
            cls.cls_behavior_attrs.register(dimension, is_dimension=True)
        # call super().__init_subclass__
        super().__init_subclass__(**kw_super)

    @classmethod
    def __setup_haver__(cls, dim_cls, **kw_super):
        '''called by dim_cls.setup_haver(cls) when setting up cls so that it "has" dim_cls dimension.'''
        cls._dim_types[cls._dimension] = dim_cls
        # call method from super if it exists
        try:
            super_setup = super().__setup_haver__
        except AttributeError:
            pass
        else:
            super_setup(dim_cls, **kw_super)

    # # # # MULTI-DIMENSION-BASED METHODS # # # #
    # # GENERIC / HELPER METHODS # #
    def dims_get(self, attr, dims=None):
        '''return dict of {dim: getattr(self.dimensions[dim], attr) for dim in dims}.
        dims: None or iterable
            if provided, only include these dimensions.
        See also: dims_apply
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        dimensions = self.dimensions if dims is None else dims
        result = dict()  # could use dict-comprehension, but explicit looping makes debugging easier.
        for d in dimensions:
            result[d] = getattr(self.dimensions[d], attr)
        return result

    def dim_values(self, dims=None):
        '''return dict of current values for dimensions in self.
        dims: None or iterable
            if provided, only include these dimensions.
        Equivalent: DimRegion(self.dims_get('v', dims=dims))
        '''
        return DimRegion(self.dims_get('v', dims=dims))

    @property
    def dims(self):
        '''return dict of current values for dimensions in self. Equivalent: self.dim_values()'''
        return self.dim_values()

    def dims_apply(self, funcname, *args_func, dims=None, **kw_func):
        '''apply funcname to each dimension in self, with args_func and kw_func.
        dims: None or iterable of strs
            if provided, only apply to these dimensions.
        See also: dims_get
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        funcs = self.dims_get(funcname, dims)
        result = dict()  # could use dict-comprehension, but explicit looping makes debugging easier.
        for d, f in funcs.items():
            result[d] = f(*args_func, **kw_func)
        return result

    # # DIMPOINTS # #
    def current_n_dimpoints(self, dims=None):
        '''return number of points represented by current values of dims.
        dims: None or iterable of strs appearing in self.dimensions.keys()
            dimensions to consider. None --> use all dimensions.

        E.g. current_n_dimpoints(self, dims=['fluid', 'snap']) --> number of (fluid, snap) points;
            e.g. 3 fluids and 2 snaps --> 6 points.

        Note, for classes using maindims, maindims are not included in the number of dimpoints.

        Equivalent to len(list(self.iter_dimpoints(dims=dims, current=True)))
        '''
        ndims = self.dims_apply('current_n', dims=dims)
        result = 1
        for n in ndims.values():
            result = result * n
        return result

    def iter_dimpoints(self, dims=None, *, all=False, restore=True, enumerate=False):
        '''iterate through values of dims, returning DimPoints and setting dim values during iteration.
        DimPoints are dicts of {dim: value} for dim in dims, where not is_iterable_dim(value).
        Also, during iteration, set self.{dim} = value, as with self.iter_dim.

        dims: None or iterable of strs appearing in self.dimensions.keys()
            dimensions to consider. None --> use all dimensions.
        all: bool
            whether to iterate through all possible values, or only the current values.
            False --> iterate through current values (e.g., self.snap, self.fluid, ...).
                    similar to itertools.product(self.iter_snap(), self.iter_fluid(), ...)
            True --> iterate through all possible values (e.g., self.snaps, self.fluid, ...)
                    similar to itertools.product(self.iter_snaps(), self.iter_fluids(), ...)
                    Equivalent to all=False if all dims are set to None, e.g. self.snap=None, ...
        restore: bool
            whether to restore original dim values after iteration.
        enumerate: bool, default False
            whether to yield indices too, i.e. (idx, DimPoint) instead of just DimPoint.
            idx would be a dict of {dim: i} such that DimPoint values are {dim: dims[i] for dim,i in idx.items()}.
        '''
        dims = self.dimensions if dims is None else dims
        applying = 'iter_values' if all else 'iter'
        iters = self.dims_apply(applying, dims=dims, enumerate=enumerate)
        # itertools.product consumes all the iters before actually iterating,
        #   so it won't set or restore values during iteration.
        with self.maintaining(**{dim: restore for dim in dims}):
            for vals in itertools.product(*iters.values()):
                if enumerate:
                    idx = {dim: val[0] for dim, val in zip(dims, vals)}
                    point = DimPoint({dim: val[1] for dim, val in zip(dims, vals)})
                    self.set_attrs(**point)
                    yield (idx, point)
                else:
                    point = DimPoint({dim: val for dim, val in zip(dims, vals)})
                    self.set_attrs(**point)
                    yield point

    def enumerate_dimpoints(self, dims=None, *, all=False):
        '''iterate through values of dims, yielding (idx, DimPoint) pairs.
        idx is a dict of {dim: i} such that DimPoint values are {dim: dims[i] for dim,i in idx.items()}.
        Also, during iteration, set self.{dim} = value, as with self.iter_dim.

        Equivalent to self.iter_dimpoints(dims=dims, all=all, enumerate=True)
        '''
        return self.iter_dimpoints(dims=dims, all=all, enumerate=True)

    def as_single_dimpoint(self, values=None, *, dims=None, **values_as_kw):
        '''return DimPoint with values for dims, but raise DimensionValueError if any value is_iterable_dim.
        values: None or dict
            values to use for the dimpoint.
                values will be joined with **values_as_kw; provided any of either will be equivalent.
                E.g. can use values={'fluid': 'e'} or use fluid='e'.
            if any are provided --> use values corresponding to self.{dim}=values[dim] for dim in dims.
            else --> use values of self.{dim} for dim in dims. (equivalent: self.dims_apply('_as_single', dims=dims)) 
        dims: None or iterable of strs appearing in self.dimensions.keys()
            dimensions to include.
            None --> infer dimensions from keys of values (and values_as_kw).
                    if no values were provided (values=None, and empty values_as_kw),
                        use all dimensions from self.dimensions.keys().

        additional kwargs provide other {dim: value} items.

        Examples:
            self.as_single_dimpoint() --> DimPoint({dim: self.{dim} for dim in self.dimensions})
            self.as_single_dimpoint({'fluid': 'e'}) --> DimPoint({'fluid': 'e'})
            self.as_single_dimpoint(fluid='e') --> DimPoint({'fluid': 'e'})
            self.as_single_dimpoint({'fluid': 'e'}, snap=0) --> DimPoint({'fluid': 'e', 'snap': 0})
            self.as_single_dimpoint(dims=['fluid', 'snap']) --> DimPoint({'fluid': self.fluid, 'snap': self.snap})
        '''
        values_dict = dict() if values is None else values
        vals_to_use = {**values_dict, **values_as_kw}
        if dims is None:  # infer from keys of vals_to_use.
            dims = list(vals_to_use.keys()) if len(vals_to_use)>0 else None
            # if dims is still None, the dims_apply method will use all dimensions from self.dimensions.keys().
        with self.using(**vals_to_use):
            result = self.dims_apply('_as_single', dims=dims)  # << raises DimensionValueError if necessary
            return DimPoint(result)

    def get_first_dimpoint(self, dims=None, *, enumerate=False):
        '''return DimPoint taking the first value of each dim in self.dimensions.
        dims: None or iterable of strs appearing in self.dimensions.keys()
            dimensions to include. None --> use all dimensions.
        enumerate: bool
            whether to return (idx, DimPoint) instead of just DimPoint.
        '''
        point = DimPoint(self.dims_apply('_get_first', dims=dims))
        if enumerate:
            idx = {dim: 0 for dim in point}
            return (idx, point)
        else:
            return point

    def using_first_dimpoint(self, dims=None):
        '''return context manager which sets dimensions to their first values (when called); restore original on exit.
        Useful for testing a single code at a single dimpoint without needing to set each dimension individually.
        dims: None or iterable of strs appearing in self.dimensions.keys()
            dimensions to include. None --> use all dimensions.
        '''
        first_point = self.get_first_dimpoint(dims=dims)
        return self.using(**first_point)

    # # ATTRS MANIPULATION # #
    def pop_dim_keys(self, kw):
        '''return ({key: kw.pop(key) for key in self.dimensions if key in kw}, kw).'''
        return ({key: kw.pop(key) for key in self.dimensions if key in kw}, kw)

    def set_pop_dim_attrs(self, kw):
        '''set self.{key} = kw.pop(key) for each key in self.dimensions if key in kw.'''
        for key in self.dimensions:
            if key in kw:
                setattr(self, key, kw.pop(key))

    def assign_dim_coords(self, array, *dims, skip=[]):
        '''assign all dimensions in self as coords for array. (self.assign_{dim}_coord(array))
        Assumes array is an xarray and does not have any dimensions in self.
        (array is not edited directly; returns result of assigning coords.)

        dims: iterable of dimensions in self
            assign only these dimensions as coords. (use all dimensions if len(dims)==0)
        skip: iterable of dimensions in self
            do not assign these dimensions as coords.
        '''
        if len(dims)==0:
            dims = self.dimensions
        if len(skip) > 0:
            dims = [dim for dim in dims if dim not in skip]
        for dim in dims:
            array = getattr(self, f'assign_{dim}_coord')(array)
        return array

    def set_attrs(self, **attrs):
        '''sets these attrs in self.'''
        for attr, val in attrs.items():
            setattr(self, attr, val)
    
    using_attrs = using_attrs
    using = alias('using_attrs')
    maintaining_attrs = maintaining_attrs
    maintaining = alias('maintaining_attrs')

    # # LOADING # #
    # multiprocessing-related properties inherited from TaskContainerCallKwargsAttrHaver; define defaults here:
    def _default_ncpu(self):
        '''return default for ncpu during self.load_across_dims. returns DEFAULTS.LOADING_NCPU'''
        return DEFAULTS.LOADING_NCPU
    def _default_timeout(self):
        '''return default for timeout during self.load_across_dims. returns DEFAULTS.LOADING_TIMEOUT'''
        return DEFAULTS.LOADING_TIMEOUT
    def _default_ncoarse(self):
        '''return default for ncoarse during self.load_across_dims. returns DEFAULTS.LOADING_NCOARSE'''
        return DEFAULTS.LOADING_NCOARSE

    # other properties for loading
    array_MBmax = simple_property('_array_MBmax', default=None, doc=_paramdocs_loading['array_MBmax'])
    @array_MBmax.getter
    def array_MBmax(self):
        '''return self._array_MBmax if it exists and is not None. Else, return DEFAULTS.ARRAY_MBYTES_MAX.'''
        result = getattr(self, '_array_MBmax', None)
        return DEFAULTS.ARRAY_MBYTES_MAX if result is None else result

    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['ncpu', 'timeout', 'ncoarse', 'print_freq', 'array_MBmax']
            + any values from super().
        '''
        kw_here = ['ncpu', 'timeout', 'ncoarse', 'print_freq', 'array_MBmax']
        kw_super = getattr(super(), '_extra_kw_for_quantity_loader_call', [])
        return kw_here + kw_super

    # helper methods for load_across_dims
    def _call_loader_at_dimpoint(self, dimpoint, loader, args_loader, **kw_loader):
        '''with self.using(**dimpoint): return loader(*args_loader, **kw_loader)'''
        # __tracebackhide__ = False, because user may with to check dimpoint (e.g. during pdb.pm())
        with self.using(**dimpoint):
            return loader(*args_loader, **kw_loader)
        # <-- even if crashed, old dims are restored at this point. dimpoint tells dims used in loader.

    def _call_loader0_at_dimpoint(self, dimpoint, loader0, args_loader, *, nload, MBmax, **kw_loader):
        '''return self._call_loader_at_dimpoint(dimpoint, loader0, ...).
        Also does memory size check based on nload and MBmax.
        intended for internal use only, inside load_across_dims.

        arr0 is handled differently from other arrays in order to do some checks.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        result = self._call_loader_at_dimpoint(dimpoint, loader0, args_loader, **kw_loader)
        memory_size_check_loading_arrays_like(result, nload=nload, MBmax=MBmax)  # crashes if too large!
        return result

    def _special_dims_shifters(self, dimnames, _shift_special):
        '''return (dict for self.using(...), dict for result.isel(...)) to _shift_special_dims.
        See docs of load_across_dims for more details.
        '''
        using = {}
        isels = {}
        for d in dimnames:
            special = _shift_special.get(d, [])
            if len(special) == 0:
                continue  # no special values specified for this dim.
            if d not in self.dimensions:
                raise InputError(f'dim {d} not in self.dimensions')
            dim = self.dimensions[d]
            if dim.current_n() <= 1:
                continue  # only 1 (or 0) values in this dim; nothing to rearrange!
            dim_v = dim.v  # <-- put this here for easier debugging.
            is_special0 = any(s == dim_v[0] for s in special)
            if not is_special0:  # [EFF] quick check for efficiency.
                continue  # first value of dim isn't special; don't need to rearrange.
            is_special = [is_special0, *[any(d == s for s in special) for d in dim_v[1:]]]
            if all(is_special):
                continue  # all values of dim are special; no point in rearranging.
            # else, rearranging can be helpful.
            first_non_special = next(i for i, s in enumerate(is_special) if not s)
            idx = [first_non_special, *range(0, first_non_special), *range(first_non_special+1, len(is_special))]
            using[d] = [dim_v[i] for i in idx]
            # need to provide isel to get back to original order...
            # e.g. if idx is [3, 0, 1, 2, 4, 5],
            #   then to go back to original order ([0,1,2,3,4,5])
            #   would need to isel result by [1,2,3,0,4,5].
            ridx = [*range(1, idx[0]+1), 0, *range(idx[0]+1, len(idx))]
            isels[d] = ridx
        return using, isels

    # load_across_dims
    @format_docstring(default_ncpu=DEFAULTS.LOADING_NCPU,
                      default_timeout=DEFAULTS.LOADING_TIMEOUT,
                      default_ncoarse=DEFAULTS.LOADING_NCOARSE,
                      default_print_freq=DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ)
    @format_docstring(**_paramdocs_loading, **_paramdocs_tasks, sub_ntab=2)
    def load_across_dims(self, loader, *args_loader, dims=[], assign_coords=None,
                         loader0=None, _shift_special={}, **kw_loader):
        '''return loader(...), iterating & joining across each dimension.

        loader: {loader}
        dims: {dims}
        assign_coords: {assign_coords}
        loader0: None or callable
            if provided, use loader0 to get the first array, then use loader for the rest.
            Internally the first array's .coords and .attrs are used to label the result;
            however all other arrays do not need to be converted to xarray.
        _shift_special: UNSET or dict of (dimstr: list of special values)
            workaround to encourage loader0 to be called on a "usual" case, not a special case.
            if provided, and dimstr in dims, and d=self.dimensions[dimstr] has multiple values,
                with special_value first, and at least one non-special value later, then
                internally rearrange dim values order before loading,
                    then rearrange result back to original order (via indexing).
            E.g. _shift_special=dict(snap=[INPUT_SNAP]) --> apply loader0 to the first non-INPUT_SNAP,
                if there are any non-INPUT_SNAP snap values in snap, and 'snap' in dims.

        --- MULTIPROCESSING STRATEGY OPTIONS (from self) ---
        timeout: {timeout}
                # [TODO] make this happen, without making self un-picklable:
                in case of crash, results so far can be found in self._latest_load_tasks.
                Then possibly continued via:
                    results = self._latest_load_tasks(..., reset=False, skip_done=True)
                    result = self._load_across_dims_postprocess(results, dims, ...)
                    # [TODO] if crashing and resuming is common, make that easier to do^
            if self.timeout has not been set, use DEFAULTS.LOADING_TIMEOUT (default: {{default_timeout}}).
        ncpu: {ncpu}
            if self.ncpu has not been set, use DEFAULTS.LOADING_NCPU (default: {{default_ncpu}}).
        ncoarse: {ncoarse}
            if self.ncoarse has not been set, use DEFAULTS.LOADING_NCOARSE (default: {{default_ncoarse}}).
        print_freq: {print_freq}
            if self.print_freq has not been set, infer from self.verbose if it exists,
            else use DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ (default: {{default_print_freq}}).

        additional args & kwargs are passed as loader(*args_loader, **kw_loader).
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if len(dims) == 0:
            return loader(*args_loader, **kw_loader)
        # else, at least 1 dim.
        dimnames = list(dims)
        dims = [(d if isinstance(d, Dimension) else self.dimensions[d]) for d in dimnames]
        # memory size check logic
        MBmax = self.array_MBmax
        nload = self.current_n_dimpoints(dims=dimnames)
        # multiprocessing strategy
        mp_strategy = dict(ncpu=self.ncpu,
                           timeout=self.timeout,
                           ncoarse=self.ncoarse,
                           print_freq=self.print_freq_explicit)
        _var = args_loader[0] if args_loader else kw_loader.get('var', None)
        _varstr = f' var={_var!r},' if isinstance(_var, str) else ''
        _printable_process_name = f'load_across_dims({loader.__name__},{_varstr} dims={dimnames}, ...)'
        # special shifters
        _shifters_using, _shifters_isels = self._special_dims_shifters(dimnames, _shift_special)
        # # METHOD 4 SETUP # #
        it_dimnames = [d.name for d in dims if d.is_iterable()]
        dshape = tuple(d.current_n() for d in dims if d.is_iterable())
        task_inputs = np.empty(dshape, dtype=object)
        # # START LOADING # #
        with contextlib.ExitStack() as context_stack:
            # set loading=(whether dim is being loaded across here), within this 'with' block.
            for d in dims:
                context_stack.enter_context(d.using(loading=True))
            unused_dims = [self.dimensions[d] for d in self.dimensions if d not in dimnames]
            for d in unused_dims:
                context_stack.enter_context(d.using(loading=False))
            # set ncpu=1 for all internal calculations (can't multiprocess inside a multiprocess)
            context_stack.enter_context(self.using(ncpu=1, pool=None))
            # restore dim values after loading, even if crash in the middle of exit block for self.using.
            #   (e.g. lots of small loading tasks + timeout error --> decent chance to crash in exit block.
            context_stack.enter_context(self.maintaining(*dimnames))
            # do special shifting
            context_stack.enter_context(self.using(**_shifters_using))
            # # LOAD ACROSS POINTS # #
            # # METHOD 4 # # -- assign all coords at the end. Use loader0 to get first array.
            #   rewritten to use TaskArray, which allows for multiprocessing.
            # setup task array
            for i, (idx, dimpoint) in enumerate(self.enumerate_dimpoints(dims=dimnames)):
                i_task = tuple(idx[d] for d in it_dimnames)  # index within tasks array
                if i == 0:
                    if loader0 is None:
                        loader0 = loader
                    task_inputs[i_task] = (self._call_loader0_at_dimpoint,
                                           (dimpoint, loader0, args_loader),
                                           {**kw_loader, 'nload': nload, 'MBmax': MBmax})
                else:
                    task_inputs[i_task] = (self._call_loader_at_dimpoint,
                                           (dimpoint, loader, args_loader),
                                           kw_loader)
            # perform tasks
            tasks = TaskArray(task_inputs, shape=dshape, printable_process_name=_printable_process_name)
            # self._latest_load_tasks = tasks   # want to do this. But, it makes self un-picklable...)
            arrs = tasks(**mp_strategy)
            result = self._load_across_dims_postprocess(arrs, dims, it_dimnames=it_dimnames,
                                                        assign_coords=assign_coords, isel=_shifters_isels)
            # # [EFF] TRYING DIFFERENT METHODS FOR EFFICIENCY # #
            # # all methods should produce equivalent results # #
            # # METHOD 1 # # -- old attempt
            #   always slower than method 2 --> removed.
            # # METHOD 2 # # -- assign final coords after all vars loaded.
            #   always slower than method 4 --> removed.
            # # METHOD 3 # # -- many assign_coord calls then xr.combine_by_coords(arrs).
            #   always slower than method 2 --> removed.
        return result

    def _load_across_dims_postprocess(self, arrs, dims, *, it_dimnames, assign_coords, isel=None):
        '''postprocessing results from load_across_dims, after all tasks have been completed.
        i.e., after all arrays have been loaded, join them all together.
        '''
        try:
            arr0 = arrs.flat[0]  # use this as the "model" array; infer coords & dims based on it.
        except IndexError:
            errmsg = 'No arrays loaded; maybe due to empty self.enumerate_dimpoints(dims=dimnames)'
            raise QuantCalcError(errmsg) from None
        # join final results
        if isinstance(arr0, xr.DataArray):
            # arrs is a numpy array (dtype=object) of xarray.DataArray objects.
            # need to convert it to a single numpy array containing all the data from each DataArray;
            # then, convert that to a single xarray.DataArray.
            # first, do some bookkeeping.
            if any(arr.shape != arr0.shape for arr in arrs.flat):
                # if shapes disagree, fix it (via broadcasting) if possible, else crash.
                arr0, arrs = self._load_across_dims_postprocess__broadcast_shapes(arr0, arrs, it_dimnames)
            # determine dims & coords for result
            dims_result = tuple([*it_dimnames, *arr0.dims])
            coords_result = arr0.coords.copy()
            for d in dims: # delete any coords that will be reassigned later!
                if d.name in coords_result:
                    del coords_result[d.name]
            # convert arrs to a single xarray.DataArray
            nparr = self._load_across_dims_postprocess__arrs_to_nparr(arr0, arrs)
            result = xr.DataArray(nparr, dims=dims_result, coords=coords_result, attrs=arr0.attrs)
            # assign coords for final result
            for d in dims:
                if d.is_iterable():
                    result = d.assign_coord_along(result, d.name)
                else:
                    result = d.assign_coord(result, overwrite=(assign_coords is not None))
        elif isinstance(arr0, xr.Dataset):
            # similar to above but handle each data var separately!
            # [TODO] encapsulate some of this stuff somewhere else?
            results = {k: [] for k in arr0.keys()}
            for k in results:
                k_arr0 = arr0[k]
                dims_k = tuple([*it_dimnames, *k_arr0.dims])
                coords_k = k_arr0.coords.copy()
                for d in dims: # delete any coords that will be reassigned later!
                    if d.name in coords_k:
                        del coords_k[d.name]
                # k_arrs is similar to arrs but for data var k, only.
                k_arrs = np.empty(arrs.shape, dtype=object)
                for idx, arr in np.ndenumerate(arrs):
                    k_arrs[idx] = arr[k]
                results[k] = xr.DataArray(k_arrs.tolist(), dims=dims_k, coords=coords_k, attrs=k_arr0.attrs)
                # assign coords for final result
                for d in dims:
                    if d.is_iterable():
                        results[k] = d.assign_coord_along(results[k], d.name)
                    else:
                        results[k] = d.assign_coord(results[k], overwrite=(assign_coords is not None))
            result = xr.Dataset(results, attrs=arr0.attrs)
        else:
            raise QuantCalcError(f'Unexpected type for arr0: {type(arr0)}; expected xarray.DataArray or Dataset.')
        if isel is not None:
            # rearrange result back to original order
            result = result.isel(isel)
        return result

    def _load_across_dims_postprocess__broadcast_shapes(self, arr0, arrs, it_dimnames):
        '''handles array broadcasting for _load_across_dims_postprocess.
        returns (arr0, arrs), but with all arrs having the same shape as arr0. Crash if not possible.
            CAUTION: may edit arrs in-place.
            (Won't edit individual arrays' data; just arrs which contains the arrays.)
        NOTE: arr0 in result isn't necessarily arrs.flat[0].
            arr0 in result is the "model array", used to model coords & dims on.
            arr0 = first DataArray with size == max(size) across all arrays.
            E.g. if array shapes are [(), (), (7,1), (1,3), (7,3), (7,3) ()],
                then use arr0 = the first DataArray with shape (7,3).
                if neither array with shape (7,3) is a DataArray (i.e., has coords...)
                then crash instead (because we can't properly assign coords to result.)
        it_dimnames: list of names of iterable dimensions; sometimes included in error message.
        '''
        shapes = [arr.shape for arr in arrs.flat]
        sizes = [arr.size for arr in arrs.flat]
        maxsize = max(sizes)
        if arr0.size != maxsize:  # there's a bigger array; use that for arr0 instead!
            for arr in arrs.flat:
                if arr.size == maxsize and isinstance(arr, xr.DataArray):
                    arr0 = arr
                    break
            else:  # didn't break... largest array is not a DataArray --> crash!
                maxshape = next(shape for arr, shape in zip(arrs.flat, shapes) if arr.size == maxsize)
                maxtypes = [type(arr) for arr in arrs.flat if arr.shape == maxshape]
                errmsg = ('array shapes are different and the largest array is not a DataArray.\n'
                          'even if broadcasting is possible, the result would be missing coord info.\n\n'
                          f'Here, largest array has shape: {maxshape}.\n'
                          f'Arrays with that shape have types: {maxtypes}.\n'
                          f'All array shapes are: {shapes}.\n\n'
                          f'Workaround: use only a single value for each dim in {it_dimnames}')
                raise DimensionalityError(errmsg)
        shape0 = arr0.shape
        # To avoid subtle errors, don't broadcast if different ndim, unless ndim=0.
        #   e.g., crash if shapes [(7, 3), (3,)] even though numpy allows broadcasting.
        #   but don't crash if shapes [(7,3), ()] or [(7,3), (1,3)] or [(7,3), (7,1)].
        # First, check which are broadcastable (crash if any incompatible)
        to_broadcast = []
        for i, shapei in enumerate(shapes):
            if shapei == shape0:
                continue
            # else, shape mismatch; handle broadcasting if possible else crash.
            if len(shapei) == 0:  # scalars are always compatible with all shapes.
                to_broadcast.append(i)
            elif len(shapei) == len(shape0):  # same ndim
                if all((si == s0 or si == 1) for si, s0 in zip(shapei, shape0)):
                    to_broadcast.append(i)  # broadcastable! :)
                else:
                    errmsg = ('load_across_dims fails when some results have different shape '
                              '(ignoring shape=() which is always compatible).\n'
                              f'Here shape={shape0} conflicts with arrs.flat[{i}].shape={shapei};\n'
                              f'All array shapes are: {shapes}.\n\n'
                              f'Workaround: use only a single value for each dim in {it_dimnames}')
                    raise DimensionalityError(errmsg)
            else:  # different ndim, and shapei is not a scalar. Crash.
                errmsg = ('load_across_dims fails when some results have different ndim; '
                          f'Here, shape={shape0} conflicts with arrs.flat[{i}].shape={shapei};\n'
                          f'All array shapes are: {shapes}.\n\n'
                          f'Workaround: use only a single value for each dim in {it_dimnames}')
                raise DimensionalityError(errmsg)
        # actually do the broadcasting:
        for i in to_broadcast:
            arrs.flat[i] = np.broadcast_to(arrs.flat[i], shape0)
        return arr0, arrs

    def _load_across_dims_postprocess__arrs_to_nparr(self, arr0, arrs):
        '''returns single numpy array (with the data from arrs, which is an array of arrays).'''
        if arr0.dtype.type is np.str_:  # special case...
            # without doing the conversion here, we get crash with:
            #    ValueError: setting an array element with a sequence
            # from call to np.asarray(arrs.tolist()) when each arr is an xr.DataArray of str(s).
            # this may be a bug with xarray itself; tested with xarray version '2024.6.0'.
            # but for now, the workaround here is fine to fix things!
            arrs_data = np.vectorize(lambda arr: arr.data)(arrs)
            arrlist = arrs_data.tolist()
        else:
            arrlist = arrs.tolist()
        # [EFF] could just do np.array(arrlist). But for large arrays, it's easy to speed up sometimes:
        if arr0.size <= 10:  # small arrays, don't do shenanigans with stack.
            nparr = np.array(arrlist)
        elif arrs.ndim == 0:
            # faster to entirely avoid conversion to numpy, when conversion not needed.
            nparr = arr0  # <-- this is an xarray.DataArray, not a numpy array, but that's okay.
        elif arrs.ndim == 1:
            # for large arrays (e.g. arr0 with >100 MB), np.stack is noticeably faster than np.array().
            nparr = np.stack(arrlist)
        elif arrs.ndim == 2:
            # for large arrays (e.g. arr0 with >100 MB), np.stack is noticeably faster than np.array().
            arrlist1d = [np.stack(aa) for aa in arrlist]
            nparr = np.stack(arrlist1d)
        else:
            # for large arrays (e.g. arr0 with >100 MB) this can get quite slow...
            #  [TODO] improve with generalized recursive calls to np.stack when ndim >= 3,
            #         continuing "pattern" from above with ndim=1 and ndim=2.
            nparr = np.asarray(arrlist)
        return nparr

    @format_docstring(**_paramdocs_loading, sub_ntab=1)
    def load_across_dims_implied_by(self, var, loader, *args_loader, assign_coords=None,
                                    _min_split=1, **kw_loader):
        '''return loader(...), iterating & joining across each dimension implied by var.
        Equivalent to self.load_across_dims(loader, ..., dims=self.match_var_loading_dims(var)).

        var: str
            variable which implies dims to load across, via self.match_var_loading_dims(var).
        loader: {loader}
        assign_coords: {assign_coords}
        _min_split: int, default 1
            if an implied dim has current_n() < min_split, don't load across it.
            1 --> no minimum.

        additional args & kwargs are passed as loader(*args_loader, **kw_loader).
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        dims = self.match_var_loading_dims(var)
        if _min_split is not None:
            dims = [d for d in dims if self.dimensions[d].current_n() >= _min_split]
        return self.load_across_dims(loader, *args_loader, dims=dims, assign_coords=assign_coords,
                                     **kw_loader)

    # # # # DISPLAY # # # #
    def _repr_contents(self):
        '''return list of contents to go in repr of self.'''
        return [f'dims={repr_simple(self.dims)}']

    def __repr__(self):
        contents = ', '.join(self._repr_contents())
        return f"{type(self).__name__}({contents})"




"""
File Purpose: helpful tools for analyzing results of instability theory

E.g. "get growth rate maxxed across k", "plot growth vs k"
"""

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .instability_xarray_accessor import itAccessor
from ...defaults import DEFAULTS
from ...errors import InputError, DimensionalityError
from ...plotting import CMAPS, current_axes_or_None
from ...quantities import angle_xy_to_hat
from ...tools import (
    UNSET,
    format_docstring,
    xarray_at_max_of,
    scalar_item,
)


''' --------------------- wavevector dims --------------------- '''

itAccessor._KNOWN_KDIMS = ['kmod_dim', 'log_kmod_dim', 'kmod', 'log_kmod', 'kang']
itAccessor.register_attr('kdims', totype=None,  # totype=None <--> works for both DataArrays & Datasets.
    value=property(lambda self: [d for d in self._KNOWN_KDIMS if d in self.obj.dims],
        doc=f'''list of dimensions associated with wavevector. E.g. ['kmod_dim', 'kang'].
        Checks: {itAccessor._KNOWN_KDIMS}'''))

itAccessor._KNOWN_KMOD_DIMS = ['kmod_dim', 'log_kmod_dim', 'kmod', 'log_kmod']
itAccessor.register_attr('kmod_dim',  # totype=None is the default, no need to specify explicitly.
    value=property(lambda self: scalar_item([d for d in self._KNOWN_KMOD_DIMS if d in self.obj.dims]),
        doc=f'''name of the dimension associated with wavevector magnitude.
        E.g. 'kmod_dim'.
        Checks: {itAccessor._KNOWN_KMOD_DIMS}'''))

itAccessor._KNOWN_KMOD_COORDS = ['kmod', 'log_kmod']
itAccessor.register_attr('kmod_coord',
    value=property(lambda self: scalar_item([c for c in self._KNOWN_KMOD_COORDS if c in self.obj.coords]),
        doc=f'''name of the coordinate associated with wavevector magnitude.
        E.g. 'kmod'.
        Checks: {itAccessor._KNOWN_KMOD_COORDS}.'''))

def _un_log10(arr):
    '''return arr but in linear space, not log10 space.
    if arr.name starts with 'log_', returns 10**arr, renamed to not have 'log_'.
    Else, returns arr, unchanged.
    '''
    return 10**arr.rename(arr.name[len('log_'):]) if arr.name.startswith('log_') else arr

itAccessor.register_attr('kmod',
    value=property(lambda self: _un_log10(self.obj[self.kmod_coord]),
        doc='''wavevector magnitude. probably array[kmod_coord]. 10**result if kmod_coord starts with 'log_'.'''))

itAccessor._KNOWN_KANG_DIMS = ['kang']
itAccessor.register_attr('kang_dim',
    value=property(lambda self: scalar_item([d for d in self._KNOWN_KANG_DIMS if d in self.obj.dims]),
        doc=f'''name of the dimension associated with wavevector angle.
        E.g. 'kang'.
        Checks: {itAccessor._KNOWN_KANG_DIMS}'''))

itAccessor.register_attr('kang',
    value=property(lambda self: self.obj[self.kang_dim] if self.kang_dim=='kang' else NotImplemented,
        doc='''wavevector angle. array['kang'].'''))

itAccessor.register_attr('has_kdims_polar',
    value=property(lambda self: any(d in self.obj.dims for d in self._KNOWN_KMOD_DIMS) \
                            and any(d in self.obj.dims for d in self._KNOWN_KANG_DIMS),
        doc='''whether the object has 2D polar kdims. I.e. one kmod_dim, and one kang_dim.'''))
itAccessor.register_attr('has_kmod_coord',
    value=property(lambda self: any(c in self.obj.coords for c in self._KNOWN_KMOD_COORDS),
        doc=f'''whether the object has a kmod coordinate. I.e. one of {itAccessor._KNOWN_KMOD_COORDS}.'''))

@itAccessor.register('khat')
def xarray_khat(array):
    '''returns unit vector in the kang direction.'''
    return angle_xy_to_hat(array.it.kang)

@itAccessor.register('k')
def xarray_k(array):
    '''returns wavevector as a vector in the x-y plane.
    (result will still have z component, but its value will be 0 everywhere.)
    '''
    return array.it.kmod * array.it.khat()


''' --------------------- growth & real(omega) --------------------- '''

def _imag_and_rename(arr):
    '''return arr.imag, with name='imag_{arr.name}'.'''
    return arr.imag.rename(f'imag_{arr.name}')

itAccessor.register_attr('growth', totype='array',
    value=property(lambda self: _imag_and_rename(self.obj) if self.obj.dtype == 'complex' else self.obj,
        doc='''growth rate. array.imag if complex, else returns array unchanged.
        if took imaginary part, result.name will be 'imag_{array.name}', instead of just array.name.'''))
itAccessor.register_attr('growth', totype='dataset',
    value=property(lambda self: _imag_and_rename(self.obj['omega']),
        doc='''growth rate. ds['omega'].rename('imag_omega').'''))

def _real_and_rename(arr):
    '''return arr.real, with name='real_{arr.name}'.'''
    return arr.real.rename(f'real_{arr.name}')

itAccessor.register_attr('omega_real', totype='array',
    value=property(lambda self: _real_and_rename(self.obj) if self.obj.dtype == 'complex' else self.obj,
        doc='''real part of omega. array.real if complex, else returns array unchanged.
        if took real part, result.name will be 'real_{array.name}', instead of just array.name.'''))
itAccessor.register_attr('omega_real', totype='dataset',
    value=property(lambda self: _real_and_rename(self.obj['omega']),
        doc='''real part of omega. ds['omega'].rename('real_omega').'''))

@itAccessor.register('at_growmax')
def xarray_at_growmax(array, growth=None):
    '''returns array where growth rate is maxxed across k dims from array.

    array: xarray.DataArray, or xarray.Dataset
        result == array.isel(location where growth is maxxed across k dims).
        if growth is None, array also tells growth rates via array.it.growth.
    growth: None, xarray.DataArray or xarray.Dataset
        None --> use array.it.growth
        DataArray --> growth rates (if real) or omega (if complex).
        Dataset --> get growth from ds['omega'].imag.
    '''
    if len(array.it.kdims)==0:  # edge case; no kdims to max across.
        return array
    if growth is None:
        growth = array.it.growth
    return xarray_at_max_of(array, of=growth, dim=growth.it.kdims)


''' --------------------- growth thinking about k --------------------- '''

itAccessor.register_attr('nonk_dims', totype='array',
    value=property(lambda self: sorted(list(set(self.obj.dims) - set(self.kdims))),
        doc='''list of dimensions associated with array but excluding kdims.
        Guaranteed to be sorted into alphabetical order.'''))
itAccessor.register_attr('nonk_dims', totype='dataset',
    value=property(lambda self: self.obj['omega'].it.nonk_dims,
        doc='''list of dimensions associated with ds['omega'] but excluding kdims.
        Guaranteed to be sorted into alphabetical order.'''))

@itAccessor.register('growth_kmax')
def xarray_growth_kmax(array):
    '''returns growth rate, maxxed across k dims.
    For DataArrays, assumes array is an array of growth rates (if real) or omega (if complex).
    For Datasets, gets growth from ds['omega'].
    '''
    return array.it.growth.max(array.it.kdims)

@itAccessor.register('grows')
def xarray_grows(array):
    '''returns boolean array telling where input growth rate (maxxed across k) > 0.
    Equivalent: xarray_growth_kmax(array) > 0.
    growth = array if real DataArray; imag(array) if complex DataArray; ds['omega'].imag if Dataset.
    '''
    return xarray_growth_kmax(array) > 0

@itAccessor.register('kmod_at_growmax')
def xarray_kmod_at_growmax(array):
    '''returns kmod where growth rate is maxxed across k.
    growth = array if real DataArray; imag(array) if complex DataArray; ds['omega'].imag if Dataset.
    '''
    return xarray_at_growmax(array.it.kmod, array.it.growth)

@itAccessor.register('kang_at_growmax')
def xarray_kang_at_growmax(array):
    '''returns kang where growth rate is maxxed across k.
    growth = array if real DataArray; imag(array) if complex DataArray; ds['omega'].imag if Dataset.
    '''
    return xarray_at_growmax(array.it.kang, array.it.growth)

@itAccessor.register('khat_at_growmax')
def xarray_khat_at_growmax(array):
    '''returns khat for kang where growth rate is maxxed across k.
    growth = array if real DataArray; imag(array) if complex DataArray; ds['omega'].imag if Dataset.
    '''
    return angle_xy_to_hat(array.it.kang_at_growmax())

@itAccessor.register('k_at_growmax')
def xarray_kds_at_growmax(array):
    '''returns dataset of kmod & kang, where growth rate is maxxed across k.
    growth = array if real DataArray; imag(array) if complex DataArray; ds['omega'].imag if Dataset.
    result is a dataset with 2 datavars (telling kmod and kang), with names:
        array.it.kmod_coord and array.it.kang_dim.
    '''
    kmod = xarray_kmod_at_growmax(array)
    kang = xarray_kang_at_growmax(array)
    return xr.Dataset({array.it.kmod_coord: kmod, array.it.kang_dim: kang})

@itAccessor.register('k_at_growmax')
def xarray_k_at_growmax(array):
    '''returns k (vector) where growth rate is maxxes across k.
    result varies with 'component', with x corresponding to kang = 0 degrees,
        and z not included (i.e., assumes k is 2D in the x-y plane).
    '''
    kmod = xarray_kmod_at_growmax(array)
    khat = xarray_khat_at_growmax(array)
    return kmod * khat


''' --------------------- wave velocity --------------------- '''

@itAccessor.register('smod_vphase')
def xarray_smod_vphase(array):
    '''returns signed magnitude of phase velocity: smod_vphase = real(omega) / |k|.
    (result's sign is same sign as real(omega).)
    Phase velocity tells wave propagation speed. Group velocity tells wave envelope speed.

    For DataArrays, assumes array is an array of real(omega) if real, or omega (if complex).
    For Datasets, gets omega from ds['omega'].

    See also: mod_vphase, vphase.
    '''
    real_omega = array.it.omega_real
    kmod = array.it.kmod
    return real_omega / kmod

@itAccessor.register('mod_vphase')
def xarray_mod_vphase(array):
    '''returns magnitude of phase velocity: mod_vphase = |real(omega)| / |k|.
    Phase velocity tells wave propagation speed. Group velocity tells wave envelope speed.

    For DataArrays, assumes array is an array of real(omega) if real, or omega (if complex).
    For Datasets, gets omega from ds['omega'].

    See also: smod_phase, vphase.
    '''
    return np.abs(array.it.smod_vphase())

@itAccessor.register('vphase')
def xarray_vphase(array):
    '''returns phase velocity (vector): vphase = (real(omega) / |k|) * khat.
    Phase velocity tells wave propagation speed. Group velocity tells wave envelope speed.

    For DataArrays, assumes array is an array of real(omega) if real, or omega (if complex).
    For Datasets, gets omega from ds['omega'].

    See also: smod_vphase, mod_vphase.
    '''
    smod_vphase = array.it.smod_vphase()
    khat = array.it.khat()
    return smod_vphase * khat


''' --------------------- targetted ops --------------------- '''

_paramdocs_targetted = {
    'growth': '''DataArray or Dataset
        xarray object containing growth info.
        real DataArray --> growth tells growth rates.
        complex DataArray --> growth tells omega; rates = omega.imag.
        Dataset --> growth rates = ds['omega'].imag''',
    'target': '''None, str, DataArray, or Dataset
        xarray object of interest.''',
}


@itAccessor.register('where_grows')
@format_docstring(**_paramdocs_targetted)
def xarray_where_grows(growth, target=None, *, drop=UNSET):
    '''return target where growth > 0. Roughly: target.where(xarray_grows(growth)).
    (or, growth.where(xarray_grows(growth)) if target is None).

    growth: {growth}
    target: {target}
        result tells target.where(grows).
    drop: UNSET or bool
        whether to drop as many nan values as possible from the result.
        UNSET --> True if growth_kmax has ndim==1, else False.
                (when growth_kmax has ndim==1, it's a list of points, easy to fully drop nans.
                when ndim>=2, might not be able to drop all the nans,
                    e.g. if (x,y)=(0,0) is nan but (0,1) and (1,0) are not, cannot drop (0,0).)

    Examples:
        dsR = xr.Dataset(...)  # dataset containing omega and maybe some other values
        dsR.it.where_grows()   # dataset where growth rate > 0 for any k. (i.e. dataset where dsR.it.grows())
        dsR.it.where_grows(drop=False)   # dataset where_grows() but keep all nan regions.
        dsR.it.where_grows(array)   # array where dsR.it.grows().
        # array where dsR.it.grows(), but assign array coords' original indexes as coords.
        #  e.g. if '_mask_stack' in array coords, result will have '_mask_stack_index' too.
        dsR.it.where_grows(array.pc.index_coords())
    '''
    if target is None:
        target = growth
    elif isinstance(target, str):
        target = growth[target]
    grows = xarray_grows(growth)
    if drop is UNSET:
        drop = (grows.ndim == 1)
    return target.where(grows, drop=drop)

@itAccessor.register('where_nogrows')
@format_docstring(**_paramdocs_targetted)
def xarray_where_nogrows(growth, target=None, *, drop=UNSET):
    '''return target where growth <= 0. Roughly: target.where(~xarray_grows(growth)).
    (or, growth.where(~xarray_grows(growth)) if target is None).

    growth: {growth}
    target: {target}
        result tells target.where(~grows).
    drop: UNSET or bool
        whether to drop as many nan values as possible from the result.

    see xarray_where_grows (or itAccessor.where_grows.f) for more details.
    '''
    if target is None:
        target = growth
    elif isinstance(target, str):
        target = growth[target]
    grows = xarray_grows(growth)
    nogrows = ~grows
    if drop is UNSET:
        drop = (nogrows.ndim == 1)
    return target.where(nogrows, drop=drop)

@itAccessor.register('stack_nonk_dims', aliases=['stack_nonk'])
@format_docstring(**_paramdocs_targetted)
def xarray_stack_nonk_dims(growth, target=None, stackdim='nonk', *, order=None):
    '''return target stacked along nonk dims.

    growth: {growth}
    target: {target}
    stackdim: str
        name of the new stacked dimension.
    order: None or list of str
        if provided, prefer dimensions in this order.
        all dims not in order will be sorted alphabetically.

    Examples:
        dsR = xr.Dataset(...)  # dataset containing omega and maybe some other values
        dsR.it.nonk_stack()   # dataset, stacking all nonk_dims
        dsR.it.nonk_stack(array)   # array, stacking all nonk_dims
    '''
    if target is None:
        target = growth
    elif isinstance(target, str):
        target = growth[target]
    nonk_dims = growth.it.nonk_dims
    if len(nonk_dims) == 0:
        errmsg = f'nonk_stack(growth,...) when growth has no nonk dims. Got growth.dims={growth.dims}'
        raise InputError(errmsg)
    if order is not None:
        ordered = [d for d in order if d in nonk_dims]
        nonk_dims = ordered + [d for d in nonk_dims if d not in ordered]
    return target.stack({stackdim: nonk_dims})


''' --------------------- plotting --------------------- '''

itAccessor.register_attr('GROWMAP', CMAPS['growlight'])
itAccessor.register_attr('GROW_VMIN', 0)
itAccessor.register_attr('GROW_VMAX', None)
itAccessor.register_attr('GROWPLOT_WRAP', 6)

@itAccessor.register('kw_growthplot')
def xarray_kw_growthplot(array, **kw_override):
    '''returns kwargs to use for a growth rate plot (using imshow-style kwargs).
    if array has_kdims_polar, use default polar=True,
        and default y=array.it.kmod_coord if it starts with 'log', else 'log_{kmod_coord}',
    if using polar=True, use default axsize of (2,3), and grid=True.
    '''
    kw = kw_override
    if array.it.has_kdims_polar:
        kw.setdefault('polar', True)
        kw.setdefault('grid', True)
        if array.it.has_kmod_coord:
            y = array.it.kmod_coord
            kw.setdefault('y', y if y.startswith('log') else f'log_{y}')
    if kw.get('polar', False):
        kw.setdefault('axsize', (2, 3))
        kw.setdefault('grid', True)
    # [TODO] how to learn what class of itAccessor we are talking about in here?
    #       right now the implementation doesn't respect if using
    #       a subclass of itAccessor which defines a different value for these.
    kw.setdefault('cmap', itAccessor.GROWMAP)
    kw.setdefault('vmin', itAccessor.GROW_VMIN)
    kw.setdefault('vmax', itAccessor.GROW_VMAX)
    return kw

@itAccessor.register('growthplots')
def xarray_growthplots(array, *, klines=None, **kw):
    '''like array.pc.subplots, but with default kwargs from kw_growthplot.
    Also, plots array.it.growth.
        (E.g. if array is complex, use array.imag instead.
        and if array is actually a Dataset, use array['omega'].imag instead.)
    Also, if array has kdims, and has only 1 nonk_dim, use defaults:
        row = the nonkdim,
        wrap = 6 if array.sizes[nonk dim] > 6.  (actually itAccessor.GROWPLOT_WRAP; default 6.)

    klines: None, bool, Dataset, dict, or tuple of (Dataset, dict)
        whether to plot lines representing mean free paths, and debye lengths, on each subplot.
        None --> True if input was a Dataset; False if it was a DataArray.
                (cannot infer klims from DataArray of growth values.)
        Dataset --> use to infer klims. Use only default kwargs for xarray_klines.
        dict --> use as kwargs for xarray_klines. infer klims from `array` input (must be a Dataset).
        tuple --> (Dataset, dict). use dataset to infer klims; use dict as xarray_klines kwargs.
        Internally, uses different kwarg defaults than xarray_klines. Defaults here:
            label = '{shortvar} ({fluid})'
            add_legend = True for axs[0,0], False for all other plots.
            log = infer from ylabel of top-left ax (if direction=='h', else xlabel of bottom left.)
    '''
    inputted = array
    array = array.it.growth
    kw = xarray_kw_growthplot(array, **kw)
    if len(array.it.kdims) > 0:
        nonk_dims = array.it.nonk_dims
        if len(nonk_dims) == 1:
            nonk_dim = nonk_dims[0]
            kw.setdefault('row', nonk_dim)
            if array.sizes[nonk_dim] > array.it.GROWPLOT_WRAP:
                kw.setdefault('wrap', array.it.GROWPLOT_WRAP)
    result = array.pc.subplots(**kw)
    # do klines
    if klines is None:
        klines = isinstance(inputted, xr.Dataset)
    if isinstance(klines, xr.Dataset):
        klines = (klines, dict())
    if isinstance(klines, tuple):
        inputted, klines = klines
    if klines != False:
        if klines == True:
            klines = dict()
        if klines.get('log', None) is None:  # infer log for all axs based on a single ax.
            if klines.get('direction', 'h') == 'h':  # use ylabel of top-left ax
                axlabel = result.axs[0,0].get_ylabel()
            else:  # use xlabel of bottom left ax.
                axlabel = result.axs[-1,0].get_xlabel()
            log = 'log' in axlabel.split('_')
            klines = {**klines, 'log': log}
        plt.sca(result.axs.flat[0])  # handle ax [0,0] separately.
        kw0 = {'add_legend': True, 'label': '{shortvar} ({fluid})',
                'legend_kw': DEFAULTS.ADDONS.GROWTHPLOTS_LEGEND_KW, **klines}
        xarray_klines(inputted.isel(result.isels.flat[0]), **kw0)
        for ax, isel, im in zip(result.axs.flat[1:], result.isels.flat[1:], result.images.flat[1:]):
            if im is None: continue  # don't draw klines on empty axes...
            plt.sca(ax)
            kwi = {'add_legend': False, 'label': '{shortvar} ({fluid})', **klines}
            xarray_klines(inputted.isel(isel), **kwi)
    return result

@itAccessor.register('growthplot')
def xarray_growthplot(array, *, klines=None, **kw):
    '''like array.pc.image, but with default kwargs from kw_growthplot.
    Also, plots array.it.growth.
        (E.g. if array is complex, use array.imag instead.
        and if array is actually a Dataset, use array['omega'].imag instead.)

    klines: None, bool, or dict
        whether to plot lines representing mean free paths, and debye lengths.
        None --> True if input was a Dataset; False if it was a DataArray.
                (cannot infer klims from DataArray of growth values.)
        if dict, use as kwargs for xarray_klines.
    '''
    inputted = array
    array = array.it.growth
    kw = xarray_kw_growthplot(array, **kw)
    result = array.pc.image(**kw)
    # do klines
    if klines is None:
        klines = isinstance(inputted, xr.Dataset)
    if klines != False:
        if klines == True:
            klines = dict()
        xarray_klines(inputted, **klines)
    return result

@itAccessor.register('klims_physical', totype='dataset', aliases=['klims_phys'])
def xarray_klims_physical(ds, *, to_simple=False):
    '''return copy of dataset, but also having the physical values limiting k.
    if ds has 'ldebye', result['kdebye'] = 2*pi/ldebye.
    if ds has 'lmfp', result['kmfp'] = 2*pi/lmfp.

    (similarly, recognizes 'eqperp_ldebye' and 'eqperp_lmfp';
        result keys would be 'eqperp_kdebye' and 'eqperp_kmfp'.)

    if result does not have anything to indicate any klims_physical in it, raise InputError.

    to_simple: bool
        whether to instead return result as a dict containing:
        (if any debye length specifiers exist in ds):
            'kdebye_lim': result[kdebye var]
            'kdebye_var': kdebye var name
        (if any mean free path specifiers exist in ds):
            'kmfp_lim': result[kmfp var]
            'kmfp_var': kmfp var name
        In this case, if multiple specifiers exist, crash.
    '''
    result = ds
    known = {'ldebye': 'kdebye',
             'lmfp': 'kmfp',
             'eqperp_ldebye': 'eqperp_kdebye',
             'eqperp_lmfp': 'eqperp_kmfp',
            }
    if all(k not in ds for k in known.keys()) and all(k not in ds for k in known.values()):
        errmsg = (f'dataset has no klims_physical. Expected any of {list(known)} or {list(known.values())};'
                  f' got dataset with keys: {list(ds.data_vars)}')
        raise InputError(errmsg)
    for lkey, kkey in known.items():
        if lkey in ds:
            kval = 2 * np.pi / ds[lkey]
            if kkey in ds and not np.all(kval == ds[kkey]):
                raise InputError(f'{kkey!r} already in ds, but does not match 2*pi/{lkey!r}.')
            result = result.assign({kkey: kval})
    if to_simple:
        dres = {}
        if 'kdebye' in result.data_vars and 'eqperp_kdebye' in result.data_vars:
            raise InputError('dataset has both kdebye and eqperp_kdebye. Cannot return as simple dict.')
        if 'kmfp' in result.data_vars and 'eqperp_kmfp' in result.data_vars:
            raise InputError('dataset has both kmfp and eqperp_kmfp. Cannot return as simple dict.')
        kdebye_var = 'kdebye' if 'kdebye' in result.data_vars else 'eqperp_kdebye'
        if kdebye_var in result.data_vars:
            dres['kdebye_lim'] = result[kdebye_var]
            dres['kdebye_var'] = kdebye_var
        kmfp_var = 'kmfp' if 'kmfp' in result.data_vars else 'eqperp_kmfp'
        if kmfp_var in result.data_vars:
            dres['kmfp_lim'] = result[kmfp_var]
            dres['kmfp_var'] = kmfp_var
        return dres
    else:
        return result

@itAccessor.register('klines', totype='dataset')
def xarray_klines(ds, *, log=None, mfp=dict(lw=2, color='blue'), debye=dict(lw=2, color='red'),
                  fluids=None, fluid_styles=UNSET, direction='h',
                  label='({val:.3g}) {log}{shortvar} ({fluid})', add_legend=True, legend_kw=UNSET,
                  **kw_line):
    '''draw lines representing mean free path, and debye length, on a plot with a |k| axis.

    ds: Dataset
        dataset containing klims_physical values. (see: xarray_klims_physical)
        (at least one of: 'ldebye', 'lmfp', 'eqperp_ldebye', 'eqperp_lmfp',
                          'kdebye', 'kmfp', 'eqperp_kdebye', 'eqperp_kmfp'.)
    log: None or bool
        whether to plot log10 of values, instead of just the values themselves.
        None --> True if plt.gca() axis label has 'log_' in it, else False.
                (ylabel if direction says axhline; else xlabel.)
    mfp: dict or False
        style to apply to mean free path lines.
        False --> don't plot mean free path lines.
    debye: dict or False
        style to apply to debye length lines.
        False --> don't plot debye length lines.
    fluids: None, 'min', 'max', or ('min', 'max')
        None --> plot lines for each fluid.
        'min' --> plot line for only 1 fluid, whichever makes line at smallest |k|
        'max' --> plot line for only 1 fluid, whichever makes line at largest |k|
        ('min', 'max') --> plot 'min' and 'max' lines.
        note: to plot an arbitrary subset of fluids, just use ds.isel(fluid=...) beforehand.
    fluid_styles: UNSET or dict of lists
        styles to ensure fluid lines appear visually distinct.
        E.g. {'ls': ['-', '--', ':']} --> use ls='-' for first fluid, '--' for next, etc.
        UNSET --> use DEFAULTS.PLOT.TIMELINES_CYCLE1; default is a dict of different linestyles.
    direction: 'h', 'v', 'horizontal', or 'vertical'
        direction to plot lines.
        'h' or 'horizontal' --> plt.axhline
        'v' or 'vertical' --> plt.axvline
    label: str
        label for each line. will be hit by .format(...) with the following kwargs:
            'log': 'log10 ' if log else ''.
            'var': variable name. 'kdebye', 'kmfp', 'eqperp_kdebye', or 'eqperp_kmfp'.
            'kvar': alias to 'var'
            'lvar': length variable name. E.g. 'ldbeye' instead of 'kdebye'.
            'pubvar': variable suitable for publication. 'k Debye' or 'k mean free path'
            'shortvar': shorthand variable name. 'debye' or 'mfp'
            'fluid': fluid name
            'val': value at which the line is plotted
    add_legend: bool
        whether to plt.legend()
    legend_kw: UNSET or dict
        kwargs to pass to plt.legend().
        UNSET --> use DEFAULTS.ADDONS.GROWTHPLOT_LEGEND_KW
    additional kwargs go to plt.axhline or plt.axvline.
    '''
    # bookkeeping
    KNOWN_DIRECTIONS = {'h': 'h', 'v': 'v', 'horizontal': 'h', 'vertical': 'v'}
    if direction not in KNOWN_DIRECTIONS:
        raise InputError(f'invalid direction: {direction!r}. Expected any of {list(KNOWN_DIRECTIONS)}')
    direction = KNOWN_DIRECTIONS[direction]
    plt_axline = plt.axhline if direction == 'h' else plt.axvline
    if log is None:
        ax = current_axes_or_None()
        if ax is not None:
            axlabel = ax.get_ylabel() if direction == 'h' else ax.get_xlabel()
            log = 'log' in axlabel.split('_')
    if isinstance(fluids, str):
        fluids = [fluids]
    if fluids is not None and not all(f=='min' or f=='max' for f in fluids):
        raise InputError(f'invalid fluids: {fluids!r}. Expected None, "min", "max", or ("min", "max")')
    nfluids = ds.sizes.get('fluid', 1) if fluids is None else len(fluids)
    if fluid_styles is UNSET:
        fluid_styles = DEFAULTS.PLOT.TIMELINES_CYCLE1
        list_fluid_styles = [{k: v[i%len(v)] for k, v in fluid_styles.items()} for i in range(nfluids)]
    # making lines -- kmfp
    kdict = xarray_klims_physical(ds, to_simple=True)
    if 'kmfp_lim' in kdict and mfp != False:
        klim = kdict['kmfp_lim']
        if log: klim = np.log10(klim)
        if 'fluid' not in klim.dims: raise NotImplementedError('[TODO] different fluid dim name?')
        if klim.ndim != 1: raise DimensionalityError(f'klines expects 1D; got ndim={klim.ndim}')
        if fluids is not None:
            fidx = []
            if 'min' in fluids: fidx.append(klim.argmin('fluid').item())
            if 'max' in fluids: fidx.append(klim.argmax('fluid').item())
            klim = klim.isel(fluid=fidx)
        for (f, lim), fstyle in zip(klim.groupby('fluid'), list_fluid_styles):
            ffmt = {
                'log': 'log10 ' if log else '',
                'var': kdict['kmfp_var'],
                'kvar': kdict['kmfp_var'],
                'lvar': kdict['kmfp_var'].replace('kmfp', 'lmfp'),
                'pubvar': 'k mean free path',
                'shortvar': 'mfp',
                'fluid': str(f),
                'val': lim.item(),
            }
            flabel = label.format(**ffmt)
            kw_here = {**kw_line, **mfp, **fstyle}  # order matters, if any keys are repeated.
            plt_axline(lim.item(), label=flabel, **kw_here)
    # making lines -- kdebye  # [TODO] encapsulate instead of copy-pasting.
    if 'kdebye_lim' in kdict and debye != False:
        klim = kdict['kdebye_lim']
        if log: klim = np.log10(klim)
        if 'fluid' not in klim.dims: raise NotImplementedError('[TODO] different fluid dim name?')
        if klim.ndim != 1: raise DimensionalityError(f'klines expects 1D; got ndim={klim.ndim}')
        if fluids is not None:
            fidx = []
            if 'min' in fluids: fidx.append(klim.argmin('fluid').item())
            if 'max' in fluids: fidx.append(klim.argmax('fluid').item())
            klim = klim.isel(fluid=fidx)
        for (f, lim), fstyle in zip(klim.groupby('fluid'), list_fluid_styles):
            ffmt = {
                'log': 'log10 ' if log else '',
                'var': kdict['kdebye_var'],
                'kvar': kdict['kdebye_var'],
                'lvar': kdict['kdebye_var'].replace('kdebye', 'ldebye'),
                'pubvar': 'k Debye',
                'shortvar': 'debye',
                'fluid': str(f),
                'val': lim.item(),
            }
            flabel = label.format(**ffmt)
            kw_here = {**kw_line, **debye, **fstyle}  # order matters, if any keys are repeated.
            plt_axline(lim.item(), label=flabel, **kw_here)
    # legend
    if add_legend:
        if legend_kw is UNSET:
            legend_kw = DEFAULTS.ADDONS.GROWTHPLOT_LEGEND_KW
        plt.legend(**legend_kw)



"""
File Purpose: Dimension values & lists for Eppic
"""
import numpy as np

from ...dimensions import Fluid, FluidList
from ...errors import InputError, InputMissingError, FluidKeyError
from ...tools import (
    alias, value_from_aliases,
    DictlikeFromKeysAndGetitem,
    format_docstring,
    UNSET,
)


''' --------------------- EppicDist and EppicDistList --------------------- '''

class EppicDist(DictlikeFromKeysAndGetitem, Fluid):
    '''info about a single distribution.
    see cls.from_dict() to create an instance from a dict of values from an eppic.i file.
    units are "raw". Numerical values equal those from the eppic.i file.

    m: real number
        mass of a single "real" particle in the distribution
    q: real number
        charge of a single "real" particle in the distribution
    n0: real number
        background "physical" number density of this distribution of particles.
    N: None, int, or str
        distribution number; will be internally stored as int (or None).
        Optional because distribution number doesn't affect distribution physics,
        however all distributions from eppic.i files will have a number.
    i: None or int
        index of this distribution in a DistributionList.
        raise NotImplementedError if i and N are incompatible
        (i.e. both non-None and representing different values).
    params: None or dict
        additional parameters.
        Provide dict with keys with trailing distN removed,
        e.g. "coll_rate" not "coll_rate1" (if dist='1')
    name: None or str
        name of this distribution.
        Optional, and probably not found in eppic.i files.
        Provide to improve human-readability.
    '''
    precision = 2  # precision when converting values to human-readable string

    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'name', 'i', 'm', 'q', 'n0', 'N', 'params'}

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, m, q, n0, *, N=None, name=None, i=None, params=None):
        N = N if N is None else int(N)
        params = dict() if params is None else params
        try:
            i_use = value_from_aliases(i=i, N=N)
        except InputError:
            # maybe eventually we will handle this case, but for now we didn't implement it.
            raise NotImplementedError(f'distribution N not equal to index in list; N={N}, i={i}')
        super().__init__(name=name, i=i_use, m=m, q=q)
        self.n0 = n0
        self.N = N
        self.params = params

    @classmethod
    def from_dict_and_N(cls, d, N):
        '''return EppicDist from dict and N.
        d: dict
            result of eppic_io_tools.read_eppic_i_file()[N].
            must have keys mdN, qdN, n0dN. E.g. "md1" for N='1'
            other keys must end in N, and one N will be removed from end of each.
                e.g. "coll_rate1" --> "coll_rate".
                raise ValueError if any key doesn't end with N.
            if 'nameN' appears in d, it will be used as name.
        N: None, str or int
            distribution number. if int, will be converted to str.
        '''
        N = str(N)
        lN = len(N)
        for key in d:
            if not key.endswith(N):
                raise ValueError(f'key {key!r} does not end with {N!r}')
        params = {key[:-lN]: value for key, value in d.items()}
        return cls(m=params['md'], q=params['qd'], n0=params['n0d'],
                   N=N, name=params.get('name', None), params=params)

    @classmethod
    def dists_from_dict(cls, d):
        '''return dict of distributions, given a dict of {N: dict of distribution values}.'''
        return {N: cls.from_dict_and_N(d, N) for N, d in d.items()}

    # # # DISPLAY # # #
    def __repr__(self):
        fmt = f'{{:.{self.precision}e}}'
        contents = []
        if self.N is not None:
            contents.append(f'N={self.N}')
        if self.name is not None:
            contents.append(f'name={self.name!r}')
        contents.extend([f'{a_}={fmt.format(getattr(self, a_))}' for a_ in ('m', 'q', 'n0')])
        content_str = ', '.join(contents)
        return f'{type(self).__name__}({content_str})'

    # # # GETTING VALUES / ITERATING # # #
    SPECIAL = ('m', 'q', 'n0', 'N', 'name')

    @format_docstring(special=SPECIAL)
    def __getitem__(self, key):
        '''usually return self.params[key]; if key is special return self.key instead.
        special keys are determined by self.SPECIAL; default {special}.
            else, return self.params[key].
        '''
        if key in self.SPECIAL:
            return getattr(self, key)
        try:
            return self.params[key]
        except KeyError:
            errmsg = f'key {key!r} not in {type(self).__name__}.params, and not special {self.SPECIAL}'
            raise KeyError(errmsg) from None

    def keys(self):
        '''return tuple of keys which are special or in self.params'''
        return tuple(key for key in (*self.SPECIAL, *self.params.keys()))

    # # # QUASINEUTRAL? # # #
    def is_hybrid(self):
        '''return whether this distribution is "hybrid", i.e. fluid model instead of PIC.
        as a proxy, guessing hybrid iff self.get('method', 0) != 0.
        '''
        return self.get('method', 0) != 0

    # # # GETTING SOME PARAMETERS # # #
    def get_n0(self):
        '''returns background (mean) density of this distribution. Equivalent: self['n0']'''
        return self['n0']

    def get_v0(self, x=None):
        '''returns original (t=0) velocity, as a tuple of (vx, vy, xz) values,
        or the x component if x provided (x should be a string if provided).
        '''
        if x is None:
            return (self.get_v0('x'), self.get_v0('y'), self.get_v0('z'))
        else:
            key = f'v{x}0d'
            return self[key]

    def get_vth0(self, x=None):
        '''returns original (t=0) thermal velocity, as a tuple of (vx, vy, xz) values,
        or the `x` component if `x` = 'x', 'y', or 'z'.

        if x is Ellipsis (i.e., x=...), assert all three components are the same,
            and return the one value which they all equal.
        '''
        if x is None:
            return (self.get_vth0('x'), self.get_vth0('y'), self.get_vth0('z'))
        elif x is ...:
            vx, vy, vz = self.get_vth0('x'), self.get_vth0('y'), self.get_vth0('z')
            assert vx == vy == vz, f'Expected all equal, but got: vth0x={vx}, vth0y={vy}, vth0z={vz}'
            return vx
        else:
            key = f'v{x}thd'
            return self[key]

    def get_nvsqr0(self, x=None):
        '''returns original (t=0) nvsqr, as a tuple of (vx^2, vy^2, vz^2) values,
        or the `x` component if `x` = 'x', 'y', or 'z'.

        The result should be comparable to eppic nvsqr output.
        To convert from known values to nvsqr...:
            Ta_x = m * (vsqr_x - u_x^2) / kB   # copied from EppicCalculator.help('Ta')
            --> vsqr_x = u_x^2 + kB * Ta_x / m.
            we know vth0_x == kB * Ta_x / m,
            and we know v0_x == u_x.
            --> nvsqr_x == n0 * (v0_x^2 + vth0_x^2)
        '''
        if x is None:
            return (self.get_nvsqr0('x'), self.get_nvsqr0('y'), self.get_nvsqr0('z'))
        else:
            return self.get_n0() * (self.get_v0(x)**2 + self.get_vth0(x)**2)

    def get_velocity_coords(self, units):
        '''
        Returns a dictionary of vdist coordinates (vx, vy, vz) for this EppicDist.
        For example, "vx": np.linspace(self["pvxmin"], self["pvxmax"], num=self["pnvx"]).
        '''
        coords = {
            "vx": np.linspace(self["pvxmin"], self["pvxmax"], num=self["pnvx"]) * units,
            "vy": np.linspace(self["pvymin"], self["pvymax"], num=self["pnvy"]) * units,
            "vz": np.linspace(self["pvzmin"], self["pvzmax"], num=self["pnvz"]) * units
        }
        return coords

    def get_npd(self, *, default=UNSET):
        '''return number of PIC particles for this distribution in each processor.
        If that is not possible (due to using alternate option for npd), raise FluidKeyError
        Alternate options include: nptotd, npcelld, nptotcelld;
            those options tell npd info but scaled by number of cells and/or processors.
            number of cells & processors are global parameters not known by EppicDist,
            so EppicDist cannot calculate npd from those options.

        default: UNSET or any value
            return default instead of error when npd not found, if default provided (not UNSET).
        '''
        npd = self.get('npd', None)
        if npd is None:
            if default is not UNSET:
                return default
            # else, raise FluidKeyError; error message depends on which alternative option in self.keys().
            keys = self.keys()
            errmsg = "'npd' not provided, and EppicDist cannot determine 'npd' from "
            if 'nptotd' in keys:
                errmsg += "'nptotd', because EppicDist doesn't know number of processors."
            elif 'npcelld' in keys:
                errmsg += "'npcelld', because EppicDist doesn't know number of cells."
            elif 'nptotcelld' in keys:
                errmsg += "'nptotcelld', because EppicDist doesn't know number of cells and processors."
            else:
                errmsg = "No indication of 'npd' was provided to EppicDist."
            raise FluidKeyError(errmsg)
        # else:
        return npd


class EppicDistList(FluidList):
    '''list of EppicDist objects'''
    value_type = EppicDist


''' --------------------- EppicNeutral and EppicNeutralList --------------------- '''

class EppicNeutral(Fluid):
    '''info about the neutral fluid, in eppic.
    [raw] units system; numerical values match input deck.

    m: number
        mass of neutral fluid particle.
    name: str, default 'neutral'
        name of the neutral fluid.
    i: int, default 0
        integer associated with the neutral fluid.
    vth: None or number
        thermal velocity of the neutral fluid. None --> unknown.
        (Note: the implementation here assumes neutral vth is isotropic.)
    '''
    precision = 2  # precision when converting values to human-readable string

    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'m', 'name', 'i', 'vth'}

    def __init__(self, m, *, name='neutral', i=0, vth=None):
        self.vth = vth
        super().__init__(name, m=m, q=0, i=i)

    # # # GETTING SOME PARAMETERS # # #
    def get_vth(self, x=None):
        '''returns thermal velocity, as a tuple of (vx, vy, xz) values,
        or the x component if x provided (x should be a string if provided).
        (Note that for neutrals, vth is the same for all 3 components.)
        if self.vth is None, raise InputMissingError.
        '''
        if self.vth is None:
            raise InputMissingError(f'cannot get_vth when {type(self).__name__}.vth is None')
        if x is None:
            return (self.vth, self.vth, self.vth)
        else:
            return self.vth

    get_vth0 = alias('get_vth',
            doc='''alias to self.get_vth. Provides interface consistency with EppicDist.get_vth0.
            note: neutral vth is the same for all 3 components, and doesn't vary with time.''')

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}(m={self.m:.{self.precision}e})'

    # # # COMPARISON # # # 
    # sometimes it's convenient to have an ordering for the values...
    # and, let's make sure EppicNeutral always compares as "less than" any EppicDist.
    # [TODO] better code encapsulation (some of this is repeated from DimensionValue class)?
    def __lt__(self, other):
        if isinstance(other, EppicDist): return True
        elif isinstance(other, type(self)): return (self.i, self.s) < (other.i, other.s)
        else: return NotImplemented
    def __le__(self, other):
        if isinstance(other, EppicDist): return True
        if isinstance(other, type(self)): return (self.i, self.s) <= (other.i, other.s)
        else: return NotImplemented
    def __gt__(self, other):
        if isinstance(other, EppicDist): return False
        if isinstance(other, type(self)): return (self.i, self.s) > (other.i, other.s)
        else: return NotImplemented
    def __ge__(self, other):
        if isinstance(other, EppicDist): return False
        if isinstance(other, type(self)): return (self.i, self.s) >= (other.i, other.s)
        else: return NotImplemented


class EppicNeutralList(FluidList):
    '''list of EppicNeutral objects.'''
    value_type = EppicNeutral

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        if len(self) > 1:
            raise NotImplementedError('EppicNeutralList with more than 1 fluid...')




"""
File Purpose: multifluid densities analysis from single-fluid muram data
"""

import numpy as np
import xarray as xr

from ...dimensions import SINGLE_FLUID
from ...defaults import DEFAULTS
from ...errors import (
    FluidValueError, FluidKeyError,
    InputError, FormulaMissingError, LoadingNotImplementedError,
)
from ...mhd import MhdMultifluidDensityLoader, Element, Specie
from ...tools import (
    simple_property,
    UNSET,
)


class MuramMultifluidDensityLoader(MhdMultifluidDensityLoader):
    '''density quantities based on Muram single-fluid values, & inferred multifluid properties.'''

    # # # N_MODE DISPATCH / CODE ARCHITECTURE # # #
    N_MODE_OPTIONS = {**MhdMultifluidDensityLoader.N_MODE_OPTIONS,
        'best': '''use best mode available, based on fluid:
            electron --> 'aux' if eosne files exist, else 'table'
            other Specie --> 'saha'.''',
        'aux': '''load directly from file. 'eosne' for electrons.
            (crash if not electron)''',
        'QN_aux': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'aux' method.
            (crash if not electron)''',
    }

    NE_MODE_OPTIONS = {**MhdMultifluidDensityLoader.NE_MODE_OPTIONS,
        'best': '''use 'aux' if eosne files exist, else 'table'.''',
        'aux': '''load directly from 'eosne' file.
            (if not possible, crash or return NaN, depending on self.typevar_crash_if_nan)''',
        'QN_aux': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'aux' method.''',
    }

    _NE_MODE_INTERNAL = {**MhdMultifluidDensityLoader._NE_MODE_INTERNAL,
        'QN_aux': 'aux',
    }

    # # # GENERIC NUMBER DENSITY # # #
    @known_var(load_across_dims=['fluid'])
    def get_ntype(self):
        '''ntype of self.fluid. Result depends on fluid as well as self.n_mode (and ne_mode if electron).
        See self.N_MODE_OPTIONS and self.NE_MODE_OPTIONS for options.
        possible results (possible in parent class)
            'SINGLE_FLUID' <--> n for SINGLE_FLUID
            'elem' <--> n for Element
            'saha' <--> n from saha equation. (not available for electrons)
            'table' <--> n from EOS table. (only available for electrons)
            'QN_table' <--> n from sum of qi ni across self.fluids, with ne from table.
        possible results (possible here but not in parent class)
            'aux' <--> n from 'eosne' file. (only available for electrons)
            'QN_aux' <--> n from sum of qi ni across self.fluids, with ne from 'eosne' file.
        '''
        result = 'nan'  # <-- nan unless set to something else below
        f = self.fluid
        if not isinstance(f, Specie):
            result = super().get_ntype()
        else:  # f is a Specie. fully handled here for readability, despite repetitiveness with super().
            # bookkeeping:
            e = f.is_electron()
            mode = self.ne_mode_explicit if e else self.n_mode
            aux = self._all_eos_aux_files_exist()
            non_aux = (not aux)
            # logic:
            if mode == 'elem' and f.element is not None:
                result = 'elem'
            elif (not e) and ((mode == 'saha') or (mode == 'best')):
                result = 'saha'
            elif e:
                if (mode == 'table') or ((mode == 'best') and non_aux):
                    result = 'table'
                elif aux and ((mode == 'aux') or (mode == 'best')):
                    result = 'aux'
                elif aux and ((mode == 'QN_aux') or (mode == 'QN')):
                    result = 'QN_aux'
                elif (mode == 'QN_table') or ((mode == 'QN') and non_aux):
                    result = 'QN_table'
        if result == 'nan':
            self._handle_typevar_nan(errmsg=f"ntype, when fluid={self.fluid!r}, n_mode={self.n_mode!r}.")
        return xr.DataArray(result)

    NTYPE_TO_VAR = {**MhdMultifluidDensityLoader.NTYPE_TO_VAR,
        'aux': 'n_aux',
    }
    _NTYPE_TO_NONSIMPLE_DEPS = {**MhdMultifluidDensityLoader._NTYPE_TO_NONSIMPLE_DEPS,
        'QN_aux': ['ne_QN', 'ne_aux'],
    }
    # _NTYPE_TO_DEPS property is inherited from MhdMultifluidDensityLoader.

    @known_var(partition_across_dim=('fluid', 'ntype'), partition_deps='_NTYPE_TO_DEPS')
    def get_n(self, *, ntype):
        '''number density. Formula depends on fluid:
        if SINGLE_FLUID, n = (r / m), from SINGLE_FLUID r & m.
            default m is the abundance-weighted average particle mass; see help(self.get_m) for details.
        if Element, n = (r / m), where
            r is inferred from abundances combined with SINGLE_FLUID r, and
            m is element particle mass (fluid.m)
        if Specie, n depends on ntype, determined from self.n_mode (and self.ne_mode, if electron);
            see help(self.get_ntype) for details.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # super() can handle the NTYPE_TO_VAR stuff, and most other stuff;
        #    here we only need to handle the new stuff in _NTYPE_TO_NONSIMPLE_DEPS.
        if ntype == 'QN_aux':
            return self('ne_QN', ne_mode='aux')
        else:
            return super().get_n(ntype=ntype)


    # # # NON-EQUILIBRIUM NUMBER DENSITY # # #

    @known_var(deps=['ne_aux'], load_across_dims=['fluid'])
    def get_n_aux(self):
        '''number density of self.fluid specie(s); from aux files values.
        Result depends on fluid:
            electron --> 'eosne'
            other --> crash with FormulaMissingError.
        '''
        # load_across_dims for this one, instead of partition_across_dim,
        #   because here we don't expect multiple self.fluid with same formula,
        #   so there's basically no efficiency improvements from grouping.
        f = self.fluid
        if f.is_electron():
            return self('ne_aux')
        else:
            raise FormulaMissingError(f'n_aux for non-electron fluid {f}.')


    # # # NTYPE: SAHA # # #
    # inherited from MhdMultifluidDensityLoader


    # # # NTYPE: ELECTRON # # #

    _NE_TYPE_TO_DEPS = {**MhdMultifluidDensityLoader._NE_TYPE_TO_DEPS,
        'aux': ['ne_aux'],
        'QN_aux': ['ne_QN', 'ne_aux'],
    }

    @known_var(value_deps=[('ne_type', '_NE_TYPE_TO_DEPS')])
    def get_ne(self):
        '''electron number density. Result depends on self.ne_mode.
        See self.NE_MODE_OPTIONS for details.
        '''
        kind = self('ne_type').item()
        # 'aux' and 'QN_aux' are new in this subclass; all other kinds can be handled by super().
        if kind == 'aux':
            return self('ne_aux')
        elif kind == 'QN_aux':
            return self('ne_QN', ne_mode='aux')
        else:
            return super().get_ne()

    # get_ne_QN() is inherited from MhdMultifluidDensityLoader.

    @known_var(dims=['snap'], ignores_dims=['fluid'])
    def get_ne_aux(self):
        '''electron number density, from 'hionne' in aux.
        hionne in aux is stored in cgs units.
        '''
        result = super().get_ne_aux()  # see MuramEosLoader
        return self._assign_electron_fluid_coord_if_unambiguous(result)



"""
File Purpose: EppicInstabilityCalculator with tfbi-specific methods
"""
import re

import numpy as np
import xarray as xr

from .eppic_instability_calculator import EppicInstabilityCalculator
from ..eppic_calculator import EppicCalculator

from ....dimensions import ELECTRON
from ....errors import (
    InputError, InputMissingError,
    DimensionalityError,
)
from ....tools import (
    simple_property,
    xarray_grid, xarray_promote_index_coords, xarray_demote_index_coords,
    xarray_log_coords,
)


class EppicTfbiCalculator(EppicInstabilityCalculator):
    '''EppicInstabilityCalculator with tfbi-specific methods.
    [TODO] some of these are generic enough to [MV] to a TfbiCalculator parent,
        if ever wanting to use a non-Eppic TfbiCalculator.
        E.g., with_chargesep_e_scaling could work for any TfbiCalculator.

    See also: EppicCalculatorWithTfbi, for an EppicCalculator which also has tfbi-related methods.
    (EppicTfbiCalculator is a FromDatasetCalculator. Not from an eppic.i file.)
    '''

    def chargesep_e_scaling(self, N=24, *, safety=0.1, name='n_mul'):
        '''returns xarray_grid of n_mul from 1 to safety * ne_at_wplasma_eq_nusn / ne.
        (Implementation assumes ne > ne_at_wplasma_eq_nusn; will crash otherwise.)

        n_at_wplasma_eq_nusn = epsilon0 nusn^2 m / q^2,
            and has aliases rosenberg_n, n_at_lmfp_eq_ldebye.

        (result * ne) spans (evenly in logspace) from ne to ne_at_wplasma_eq_nusn.

        N: int
            number of points in result
        name: str
            name of resulting array and coordinate.
        safety: number, probably less than 1
            safety factor for the range of n_mul.
            smaller safety is MORE safe (extending the search into more drastic n_mul).
        '''
        ne = self('n', fluid=ELECTRON)
        ne_at_wplasma_eq_nusn = self('n_at_wplasma_eq_nusn', fluid=ELECTRON)
        rat = ne_at_wplasma_eq_nusn/ne
        if rat.size == 1:     # if rat can be converted to a single scalar, do that,
            rat = rat.item()  #   to avoid using array_lims mode (which includes 'n_mul_dim').
        result = xarray_grid(safety*rat, 1, N, name=name, logspace=True, reverse=True)
        return result

    def with_chargesep_e_scaling(self, N=24, safety=0.1, **kw_init):
        '''returns self with_scaling all n by grid from 1 to safety * ne_at_wplasma_eq_nusn / ne.
        solve tfbi across result to evaluate the "okay-ness" of scaling all number densities.
        At some point in this range, will probably see significant changes in tfbi solution.

        smaller safety is MORE safe (extending the search into more drastic n_mul).

        Equivalent: self.with_scaling({'n': self.chargesep_e_scaling(...)})
        '''
        chargesep_e_scaling = self.chargesep_e_scaling(N, safety=safety, name='n_mul')
        return self.with_scaling({'n': chargesep_e_scaling}, **kw_init)


class EppicCalculatorWithTfbi(EppicCalculator):
    '''EppicCalculator which knows some tfbi-related methods.
    see help(EppicCalculator) for more details.
    '''

    # # # BASES # # #
    def _get_n_neutral(self):
        '''get the number density of self.fluid when it is an EppicNeutral or EppicNeutralList.
        (Or, crash with a helpful FormulaMissingError error message.)

        The implementation here returns the value of _coord_ngrid,
            either from eppic.i file, or from self.tfbi_grid1_zeros if available.
            (Both should be very similar, but value from eppic.i file has only 4 sigfigs.)
        Does NOT check super()._get_n_neutral().
            Setting self.set_var('nj', value, ...) is untested; might not do anything.

        result will be a xarray.DataArray with ndim=0.
        '''
        input_coord = self.input_deck['_coord_ngrid']
        if self.tfbi_grid1_zeros is None:
            result = input_coord
        else: # get the more precise value
            index = self.input_deck['_coord_ngrid_index']
            result = self.tfbi_grid1_zeros['ngrid'].isel(ngrid=index).item()
        return xr.DataArray(result)


    # # # INDEX1 STUFF # # #

    INDEX1_KEYS = ('EBspeed', 'Tgrid', 'ionfrac_H', 'kappae', 'ngrid')

    cls_behavior_attrs.register('tfbi_grid1_zeros', default=None)
    tfbi_grid1_zeros = simple_property('_tfbi_grid1_zeros', default=None,
        doc='''5D grid of zeros including all possibilities for tfbi_index1 coordinates.
        Must be provided separately in order to compute self('tfbi_index1_delta'),
            the array of all zeros EXCEPT at the point corresponding to index1 of this calculator.
        Example:
            DSR_DIR = 'Grid01.pcxarr'
            dsR00 = pc.xarray_load(DSR_DIR).load().drop_vars('T')  # T==Tgrid; don't keep both
            dsR0 = dsR00.pc.index_coords(max_ndim=1)  # add the index coords into the grid
            self.tfbi_grid1_zeros = xr.zeros_like(dsR0.it.growth_kmax())''')

    @known_var
    def get_tfbi_index1_with_eppici_coords(self):
        '''tfbi index1 value as a size=1 DataArray with 5-tuple as its element, and dtype=object.
        The index1 tells indexes from "grid01" for (in this order):
            EBspeed, Tgrid, ionfrac_H, kappae, ngrid

        indexes come from the input deck, from param names like '_coord_EBspeed_index'.

        Result has shape (1,1,1,1,1). Result also has coords for each key.
        '''
        KEYS = self.INDEX1_KEYS
        indexes = {k: self.input_deck[f'_coord_{k}_index'] for k in KEYS}
        cvalues = {k: self.input_deck[f'_coord_{k}'] for k in KEYS}  # approximate; has rounding errors.
        indexes_coords = {f'{k}_index': (k, [index]) for k, index in indexes.items()}
        cvalues_coords = {k: (k, [cval]) for k, cval in cvalues.items()}
        index1 = tuple(indexes[k] for k in KEYS)
        result = np.array([[[[[None]]]]], dtype=object)
        result[0,0,0,0,0] = index1
        return xr.DataArray(result, dims=KEYS, coords={**indexes_coords, **cvalues_coords})

    @known_var(deps=['tfbi_index1_with_eppici_coords'])
    def get_tfbi_index1(self):
        '''tfbi index1 value as a size=1 DataArray with 5-tuple as its element, and dtype=object.
        The index1 tells indexes from "grid01" for (in this order):
            EBspeed, Tgrid, ionfrac_H, kappae, ngrid
        The dims of result will {k}_index, e.g. 'EBspeed_index'.

        Result has shape (1,1,1,1,1). Result does NOT also have coords for each key.
        '''
        result = self('tfbi_index1_with_eppici_coords')
        result = xarray_promote_index_coords(result)
        return result.drop_vars(self.INDEX1_KEYS)

    @known_var(deps=['tfbi_index1_with_eppici_coords'])
    def get_tfbi_index1_with_coords(self):
        '''tfbi index1 value as a size=1 DataArray with 5-tuple as its element, and dtype=object.
        The index1 tells indexes from "grid01" for (in this order):
            EBspeed, Tgrid, ionfrac_H, kappae, ngrid

        indexes come from the input deck, from param names like '_coord_EBspeed_index'.
        Result has shape (1,1,1,1,1). Result also has coords for each key.

        Coords for each key come from the input deck UNLESS self.tfbi_grid1_zeros is available,
        in which case the more precise values from tfbi_grid1_zeros are used.
        (e.g. coord from eppic.i might say 3.162, while tfbi_grid1_zeros might say 3.162277660)
        '''
        result = self('tfbi_index1_with_eppici_coords')
        if self.tfbi_grid1_zeros is not None:
            result = self.assign_tfbi_index1_coords(result)
        return result

    @known_var(deps=['tfbi_index1_with_coords'])
    def get_tfbi_index1_with_log_coords(self):
        '''tfbi index1 value as a size=1 DataArray with 5-tuple as its element, and dtype=object.
        Like tfbi_index1_with_coords, but also has 'log_{key}' coords for easy reference.
        i.e., result has log_EBspeed, log_Tgrid, log_ionfrac_H, log_kappae, and log_ngrid coords, too.
        '''
        result = self('tfbi_index1_with_coords')
        return xarray_log_coords(result, self.INDEX1_KEYS, promote=False, drop=False)

    @known_var(deps=['tfbi_index1'])
    def get_tfbi_index1_bool(self):
        '''tfbi index1 value as a size=1 DataArray([[[[[True]]]]]),
        with "grid01" dims each having size=1, and including coords & index coords for each:
            EBspeed, Tgrid, ionfrac_H, kappae, ngrid
        '''
        return xr.ones_like(self('tfbi_index1'), dtype=bool)

    @known_var
    def get_tfbi_grid1_zeros(self):
        '''5D grid of zeros including all possibilities for tfbi_index1 coordinates.
        Internally stored (and can be altered) via self.tfbi_grid1_zeros.
        '''
        if self.tfbi_grid1_zeros is None:
            errmsg = ('self.tfbi_grid1_zeros must be provided (i.e. not None)\n'
                      'See help(type(self)).tfbi_grid1_zeros for more details.')
            raise InputMissingError(errmsg)
        return self.tfbi_grid1_zeros

    @known_var(deps=['tfbi_index1_bool', 'tfbi_grid1_zeros'])
    def get_tfbi_index1_delta(self):
        '''5D grid of zeros EXCEPT at the index1 for this calculator.
        The 5D grid includes all possible coordinate combinations for tfbi_index1.
        '''
        i1 = self('tfbi_index1_bool')
        grid1 = self('tfbi_grid1_zeros')
        grid1 = xarray_promote_index_coords(grid1)  # align by index; coord values include rounding errors.
        result = i1.reindex_like(grid1, fill_value=0)
        result = result + grid1   # include original grid1 coord values (for non-index coords)
        result = xarray_demote_index_coords(result)
        return result.astype(bool)

    def tfbi_index1_from_str(self, s, *, to='bool'):
        '''determine tfbi index1 from str like "index1=(1,2,3,4,5)" or just "(1,2,3,4,5)".
        
        to: 'tuple', 'xrtuple', 'bool', or 'delta'
            tuple --> just return 5-tuple of ints.
            xrtuple --> result is formatted like self('tfbi_index1') result.
                    i.e., size=1 DataArray with 5-tuple as its element, and dtype=object.
                    result dims will be {key}_index, e.g. 'EBspeed_index'.
            bool --> result is formatted like self('tfbi_index1_bool') result.
                    i.e., size=1 DataArray([[[[[True]]]]]), with all 5 index1 dims.
                    result dims will be {key}_index, e.g. 'EBspeed_index'.
            delta --> result is formatted like self('tfbi_index1_delta') result.
                    i.e., 5D grid of zeros EXCEPT at the index1 for this calculator.
                    result dims will be {key}, e.g. 'EBspeed'.
                    requires non-None self.tfbi_grid1_zeros.
        '''
        i = r'\s*(\d+)\s*'
        pattern = rf'.*index1\s*=\s*\({i},{i},{i},{i},{i}\).*'
        match = re.match(pattern, s)
        if match is None:
            raise InputError(f'input not like "index1=(1,2,3,4,5)". Got: {s!r}')
        indexes_from_str = match.groups()
        index1 = tuple(int(i) for i in indexes_from_str)
        if to == 'tuple':
            return index1
        # [TODO] encapsulate repeated code from earlier in this class.
        KEYS = self.INDEX1_KEYS
        indexes = dict(zip(KEYS, index1))
        indexes_coords = {f'{k}_index': (f'{k}_index', [index]) for k, index in indexes.items()}
        result = np.array([[[[[None]]]]], dtype=object)
        result[0,0,0,0,0] = index1
        result = xr.DataArray(result, dims=indexes_coords.keys(), coords=indexes_coords)
        if to == 'xrtuple':
            return result
        result = xr.ones_like(result, dtype=bool)
        if to == 'bool':
            return result
        if to != 'delta':
            errmsg = (f'unknown to={to!r}. Expected "tuple", "xrtuple", "bool", or "delta".')
            raise InputError(errmsg)
        # to == delta:
        grid1 = self('tfbi_grid1_zeros')
        grid1 = xarray_promote_index_coords(grid1)
        result = result.reindex_like(grid1, fill_value=0)
        result = result + grid1
        result = xarray_demote_index_coords(result)
        return result.astype(bool)

    def assign_tfbi_index1_coords(self, array):
        '''return array after assigning any index-related coords from tfbi_index1,
        based on values from self.tfbi_grid1_zeros, and index coords in array.

        This method expects array to have 'index' coords (possibly as dims),
            and overwrites the associated non-index coords.
            E.g. array has EBspeed_index --> result's EBspeed coord comes from grid1.
        '''
        grid1 = self('tfbi_grid1_zeros')
        for k in self.INDEX1_KEYS:
            k_index = f'{k}_index'
            if k_index not in array.coords:
                continue
            gdim = grid1.coords[k].dims[0]
            gcvalk = grid1.coords[k]
            acvalk = array.coords[k_index]
            if acvalk.ndim == 0:  # scalar coord, don't need to worry about array dims
                to_assign = gcvalk.isel({gdim: acvalk.item()}).values
                array = array.assign_coords({k: to_assign})
            elif acvalk.ndim == 1:  # need to determine which array dim has k_index coord
                to_assign = gcvalk.isel({gdim: acvalk.values}).values
                adim = acvalk.dims[0]
                array = array.assign_coords({k: (adim, to_assign)})
            else:  # acvalk.ndim >= 2
                errmsg = f'assign_tfbi_index1_coords when array[{k_index!r}].ndim > 1'
                raise DimensionalityError(errmsg)
        return array



"""
File Purpose: base quantities for EbysusCalculator
"""
import xarray as xr

from ...quantities import AllBasesLoader, SimpleDerivedLoader

class EbysusBasesLoader(AllBasesLoader, SimpleDerivedLoader):
    '''base quantities based on Ebysus output.'''

    # # # BASE_QUANTS which are properties of Ebysus # # #

    @known_var
    def get_gamma(self):
        '''adiabatic index.'''
        return 5/3   # it's always 5/3 for ebysus... other values remain untested.


    # # # BASE_QUANTS directly from Ebysus # # #
    
    @known_var(load_across_dims=['fluid'])
    def get_m(self):
        '''mass.'''
        return xr.DataArray(self.u('amu') * self.fluid.m, attrs=dict(units=self.units))

    @known_var(load_across_dims=['fluid'])
    def get_q(self):
        '''charge. (directly from Ebysus)'''
        return xr.DataArray(self.u('qe') * self.fluid.q, attrs=dict(units=self.units))

    @known_var(load_across_dims=['component'], aliases=['dspace', 'ds_sim'])
    def get_ds(self):
        '''vector(spatial scale), e.g. [dx, dy, dz]. Depends on self.component.'''
        x = str(self.component)
        dx = getattr(self.dd, f'd{x}') * self.dd.uni('l')
        return xr.DataArray(dx, attrs=dict(units=self.units))


    # # # BASE_QUANTS loaded directly # # #

    @known_var(dims=['snap', 'component'])
    def get_B(self):
        '''magnetic field. (directly from Ebysus)'''
        return self.load_maindims_var_across_dims('b', dims=['snap', 'component'])


    # # # other ebysus direct inputs # # #

    @known_var(dims=['snap', 'fluid'])
    def get_r(self):
        '''mass density. (directly from Ebysus)'''
        return self.load_maindims_var_across_dims('ri', dims=['snap', 'fluid'])

    @known_var(dims=['snap', 'fluid', 'component'])
    def get_p(self):
        '''momentum density. (directly from Ebysus)'''
        return self.load_maindims_var_across_dims('pi', dims=['snap', 'fluid', 'component'])


    # # # BASE_QUANTS derived from ebysus  # # #

    @known_var(deps=['r', 'm'])
    def get_n(self):
        '''number density. n = (r / m) = (mass density / mass)'''
        return self('r') / self('m')

    @known_var(deps=['p', 'r'])
    def get_u(self):
        '''velocity. u = p / r = (momentum density / mass density)'''
        return self('p') / self('r')


    # # # BASE_QUANTS calculated by helita, from ebysus  # # #
    # [TODO] calculate these in PlasmaCalcs instead.
    #    check PlasmaCalcs results against the helita calculations, for accuracy;
    #    check whether PlasmaCalcs or helita is faster.
    #    E.g. get_J_helita vs get_J, defined below.

    @known_var(dims=['snap', 'fluid'])
    def get_T(self):
        '''temperature. (directly from Ebysus).
        "maxwellian" temperature (classical T in thermodynamics).
        '''
        return self.load_maindims_var_across_dims('tg', dims=['snap', 'fluid'])

    # def get_nusj(self):
    #   defined in ebysus_collisions.py, instead.

    # def get_E(self):
    #   defined in ebysus_efield.py, instead.


    # # # CURRENT DENSITY (helita & here)  # # #

    @known_var(dims=['snap', 'component'])
    def get_J_helita(self):
        '''current density. (directly from Ebysus)'''
        return self.load_maindims_var_across_dims('j', dims=['snap', 'component'])

    @known_var(load_across_dims=['component'], aliases=['J_imposed'])
    def get_J_ext(self):
        '''imposed current density; see ic_ix, ic_iy, ic_iz from Ebysus.'''
        if self.dd.get_param('do_imposed_current', 0) <= 0:
            return self('0')  # not imposing any current.
        x = str(self.component)
        ic_units = self.dd.get_param('ic_units', 'ebysus').strip().lower()
        if ic_units == 'ebysus': ic_units = 'simu'
        result = self.dd.get_param(f'ic_i{x}', 0) * self.dd.uni('i', units_input=ic_units)
        return xr.DataArray(result, attrs=dict(units=self.units))

    @known_var(deps=['curl_B'])
    def get_J_B(self):
        '''current density from magnetic field (ignoring J_ext). J_B = curl(B) / mu0.'''
        return self('curl_B') / self.dd.uni('mu0')

    @known_var(deps=['J_B', 'J_ext'])
    def get_J(self):
        '''total current density. J == J_B + J_ext == curl(B) / mu0 + J_imposed.'''
        # [EFF] (2024/03/16) faster than J_helita when getting 10+ snaps with 3 components.
        #     Slower than J_helita when getting only 1 snap or only 1 component.
        #     (within a factor of 2x speed of J_helita either way.)
        # accuracy check: seems to be reasonably similar when using dd.stagger_kind='first';
        #     also similar (but more outliers) when using dd.stagger_kind='fifth'.
        return self('J_B') + self('J_ext')
